# coding: utf-8

from __future__ import division

__status__ = "Development"

import logging
import logging.config
import os
import unittest

from doped.pycdt.utils.log_util import initialize_logging


class InitializeLoggingTest(unittest.TestCase):
    def test_initialize_logging(self):
        self.assertEqual(initialize_logging(), None)
        self.assertEqual(initialize_logging(filename="this_tmp_filename.txt"), None)
        os.remove("this_tmp_filename.txt")
        self.assertEqual(initialize_logging(level="WARNING"), None)
        self.assertEqual(initialize_logging(level="DEBUG"), None)
        self.assertEqual(initialize_logging(level="NOTSET"), None)


import unittest

if __name__ == "__main__":
    unittest.main()



import os
import unittest
import warnings
from unittest.mock import patch

import numpy as np
from pymatgen.core.structure import Structure

from doped.pycdt.utils import parse_calculations


class DopedParsingTestCase(unittest.TestCase):
    def setUp(self):
        # get module path
        self.module_path = os.path.dirname(os.path.abspath(__file__))
        self.EXAMPLE_DIR = os.path.join(self.module_path, "../examples")
        self.CDTE_EXAMPLE_DIR = os.path.join(self.module_path, "../examples/CdTe")
        self.CDTE_BULK_DATA_DIR = os.path.join(
            self.CDTE_EXAMPLE_DIR, "Bulk_Supercell/vasp_ncl"
        )
        self.cdte_dielectric = np.array(
            [[9.13, 0, 0], [0.0, 9.13, 0], [0, 0, 9.13]]
        )  # CdTe

        self.ytos_dielectric = [
            [40.71948719643814, -9.282128210266565e-14, 1.26076160303219e-14],
            [-9.301652644020242e-14, 40.71948719776858, 4.149879443489052e-14],
            [5.311743673463141e-15, 2.041077680836527e-14, 25.237620491130023],
        ]
        # from legacy Materials Project

    def test_dielectric_initialisation(self):
        """
        Test that dielectric can be supplied as float or int or 3x1 array/list or 3x3 array/list
        """
        defect_file_path = f"{self.CDTE_EXAMPLE_DIR}/vac_1_Cd_-2/vasp_ncl"
        # test float
        sdp = parse_calculations.SingleDefectParser.from_paths(
            path_to_defect=defect_file_path,
            path_to_bulk=self.CDTE_BULK_DATA_DIR,
            dielectric=9.13,
            defect_charge=-2,
        )
        self.assertTrue(
            np.allclose(sdp.defect_entry.parameters["dielectric"], 9.13)
        )  # # float/int stay the same

        # test int
        sdp = parse_calculations.SingleDefectParser.from_paths(
            path_to_defect=defect_file_path,
            path_to_bulk=self.CDTE_BULK_DATA_DIR,
            dielectric=9,
            defect_charge=-2,
        )
        self.assertTrue(
            np.allclose(
                sdp.defect_entry.parameters["dielectric"],
                9,  # float/int stay the same
            )
        )

        # test 3x1 array
        sdp = parse_calculations.SingleDefectParser.from_paths(
            path_to_defect=defect_file_path,
            path_to_bulk=self.CDTE_BULK_DATA_DIR,
            dielectric=np.array([9.13, 9.13, 9.13]),
            defect_charge=-2,
        )
        self.assertTrue(
            np.allclose(sdp.defect_entry.parameters["dielectric"], self.cdte_dielectric)
        )

        # test 3x1 list
        sdp = parse_calculations.SingleDefectParser.from_paths(
            path_to_defect=defect_file_path,
            path_to_bulk=self.CDTE_BULK_DATA_DIR,
            dielectric=[9.13, 9.13, 9.13],
            defect_charge=-2,
        )
        self.assertTrue(
            np.allclose(sdp.defect_entry.parameters["dielectric"], self.cdte_dielectric)
        )

        # test 3x3 array
        sdp = parse_calculations.SingleDefectParser.from_paths(
            path_to_defect=defect_file_path,
            path_to_bulk=self.CDTE_BULK_DATA_DIR,
            dielectric=self.cdte_dielectric,
            defect_charge=-2,
        )
        self.assertTrue(
            np.allclose(sdp.defect_entry.parameters["dielectric"], self.cdte_dielectric)
        )

        # test 3x3 list
        sdp = parse_calculations.SingleDefectParser.from_paths(
            path_to_defect=defect_file_path,
            path_to_bulk=self.CDTE_BULK_DATA_DIR,
            dielectric=self.cdte_dielectric.tolist(),
            defect_charge=-2,
        )
        self.assertTrue(
            np.allclose(sdp.defect_entry.parameters["dielectric"], self.cdte_dielectric)
        )

    def test_vacancy_parsing_and_freysoldt(self):
        """Test parsing of Cd vacancy calculations and correct Freysoldt correction calculated"""
        parsed_vac_Cd_dict = {}

        for i in os.listdir(self.CDTE_EXAMPLE_DIR):
            if "vac_1_Cd" in i:  # loop folders and parse those with "vac_1_Cd" in name
                defect_file_path = f"{self.CDTE_EXAMPLE_DIR}/{i}/vasp_ncl"
                defect_charge = int(i[-2:].replace("_", ""))
                # parse with no transformation.json:
                sdp = parse_calculations.SingleDefectParser.from_paths(
                    path_to_defect=defect_file_path,
                    path_to_bulk=self.CDTE_BULK_DATA_DIR,
                    dielectric=self.cdte_dielectric,
                    defect_charge=defect_charge,
                )
                sdp.freysoldt_loader()
                sdp.get_stdrd_metadata()
                sdp.get_bulk_gap_data()
                sdp.run_compatibility()
                parsed_vac_Cd_dict[
                    i
                ] = sdp.defect_entry  # Keep dictionary of parsed defect entries

        self.assertTrue(len(parsed_vac_Cd_dict) == 3)
        self.assertTrue(all(f"vac_1_Cd_{i}" in parsed_vac_Cd_dict for i in [0, -1, -2]))
        # Check that the correct Freysoldt correction is applied
        for name, energy, correction_dict in [
            (
                "vac_1_Cd_0",
                4.166,
                {"bandfilling_correction": -0.0, "bandedgeshifting_correction": 0.0},
            ),
            (
                "vac_1_Cd_-1",
                6.355,
                {
                    "charge_correction": 0.22517150393292082,
                    "bandfilling_correction": -0.0,
                    "bandedgeshifting_correction": 0.0,
                },
            ),
            (
                "vac_1_Cd_-2",
                8.398,
                {
                    "charge_correction": 0.7376460317828045,
                    "bandfilling_correction": -0.0,
                    "bandedgeshifting_correction": 0.0,
                },
            ),
        ]:
            self.assertAlmostEqual(parsed_vac_Cd_dict[name].energy, energy, places=3)
            for correction_name, correction_energy in correction_dict.items():
                self.assertAlmostEqual(
                    parsed_vac_Cd_dict[name].corrections[correction_name],
                    correction_energy,
                    places=3,
                )

            # assert auto-determined vacancy site is correct
            # should be: PeriodicSite: Cd (6.5434, 6.5434, 6.5434) [0.5000, 0.5000, 0.5000]
            if name == "vac_1_Cd_0":
                np.testing.assert_array_almost_equal(
                    parsed_vac_Cd_dict[name].site.frac_coords, [0.5, 0.5, 0.5]
                )
            else:
                np.testing.assert_array_almost_equal(
                    parsed_vac_Cd_dict[name].site.frac_coords, [0, 0, 0]
                )

    def test_interstitial_parsing_and_kumagai(self):
        """Test parsing of Te (split-)interstitial and Kumagai-Oba (eFNV) correction"""
        with patch("builtins.print") as mock_print:
            for i in os.listdir(self.CDTE_EXAMPLE_DIR):
                if "Int_Te" in i:  # loop folders and parse those with "Int_Te" in name
                    defect_file_path = f"{self.CDTE_EXAMPLE_DIR}/{i}/vasp_ncl"
                    defect_charge = int(i[-2:].replace("_", ""))
                    # parse with no transformation.json:
                    sdp = parse_calculations.SingleDefectParser.from_paths(
                        path_to_defect=defect_file_path,
                        path_to_bulk=self.CDTE_BULK_DATA_DIR,
                        dielectric=self.cdte_dielectric,
                        defect_charge=defect_charge,
                    )
                    sdp.kumagai_loader()
                    sdp.get_stdrd_metadata()
                    sdp.get_bulk_gap_data()
                    sdp.run_compatibility()
                    te_i_2_ent = sdp.defect_entry

        mock_print.assert_called_once_with(
            "Saving parsed Voronoi sites (for interstitial site-matching) "
            "to bulk_voronoi_sites.json to speed up future parsing."
        )

        self.assertAlmostEqual(te_i_2_ent.energy, -6.221, places=3)
        self.assertAlmostEqual(te_i_2_ent.uncorrected_energy, -7.105, places=3)
        correction_dict = {
            "charge_correction": 0.8834518111049584,
            "bandfilling_correction": -0.0,
            "bandedgeshifting_correction": 0.0,
        }
        for correction_name, correction_energy in correction_dict.items():
            self.assertAlmostEqual(
                te_i_2_ent.corrections[correction_name], correction_energy, places=3
            )

        # assert auto-determined interstitial site is correct
        # should be: PeriodicSite: Te (12.2688, 12.2688, 8.9972) [0.9375, 0.9375, 0.6875]
        np.testing.assert_array_almost_equal(
            te_i_2_ent.site.frac_coords, [0.9375, 0.9375, 0.6875]
        )

        # run again to check parsing of previous Voronoi sites
        with patch("builtins.print") as mock_print:
            for i in os.listdir(self.CDTE_EXAMPLE_DIR):
                if "Int_Te" in i:  # loop folders and parse those with "Int_Te" in name
                    defect_file_path = f"{self.CDTE_EXAMPLE_DIR}/{i}/vasp_ncl"
                    defect_charge = int(i[-2:].replace("_", ""))
                    # parse with no transformation.json:
                    sdp = parse_calculations.SingleDefectParser.from_paths(
                        path_to_defect=defect_file_path,
                        path_to_bulk=self.CDTE_BULK_DATA_DIR,
                        dielectric=self.cdte_dielectric,
                        defect_charge=defect_charge,
                    )
                    sdp.kumagai_loader()
                    sdp.get_stdrd_metadata()
                    sdp.get_bulk_gap_data()
                    sdp.run_compatibility()
                    te_i_2_ent = sdp.defect_entry

        mock_print.assert_not_called()
        os.remove("bulk_voronoi_nodes.json")

    def test_substitution_parsing_and_kumagai(self):
        """Test parsing of Te_Cd_1 and Kumagai-Oba (eFNV) correction"""
        for i in os.listdir(self.CDTE_EXAMPLE_DIR):
            if "as_1_Te" in i:  # loop folders and parse those with "as_1_Te" in name
                defect_file_path = f"{self.CDTE_EXAMPLE_DIR}/{i}/vasp_ncl"
                defect_charge = int(i[-2:].replace("_", ""))
                # parse with no transformation.json:
                sdp = parse_calculations.SingleDefectParser.from_paths(
                    path_to_defect=defect_file_path,
                    path_to_bulk=self.CDTE_BULK_DATA_DIR,
                    dielectric=self.cdte_dielectric,
                    defect_charge=defect_charge,
                )
                sdp.kumagai_loader()
                sdp.get_stdrd_metadata()
                sdp.get_bulk_gap_data()
                sdp.run_compatibility()
                te_cd_1_ent = sdp.defect_entry

        self.assertAlmostEqual(te_cd_1_ent.energy, -2.665996, places=3)
        self.assertAlmostEqual(te_cd_1_ent.uncorrected_energy, -2.906, places=3)
        correction_dict = {
            "charge_correction": 0.24005014473002428,
            "bandfilling_correction": -0.0,
            "bandedgeshifting_correction": 0.0,
        }
        for correction_name, correction_energy in correction_dict.items():
            self.assertAlmostEqual(
                te_cd_1_ent.corrections[correction_name], correction_energy, places=3
            )
        # assert auto-determined substitution site is correct
        # should be: PeriodicSite: Te (6.5434, 6.5434, 6.5434) [0.5000, 0.5000, 0.5000]
        np.testing.assert_array_almost_equal(
            te_cd_1_ent.site.frac_coords, [0.5000, 0.5000, 0.5000]
        )

    def test_extrinsic_interstitial_defect_ID(self):
        """Test parsing of extrinsic F in YTOS interstitial"""
        bulk_sc_structure = Structure.from_file(f"{self.EXAMPLE_DIR}/YTOS/Bulk/POSCAR")
        initial_defect_structure = Structure.from_file(
            f"{self.EXAMPLE_DIR}/YTOS/Int_F_-1/Relaxed_CONTCAR"
        )
        (
            def_type,
            comp_diff,
        ) = parse_calculations.get_defect_type_and_composition_diff(
            bulk_sc_structure, initial_defect_structure
        )
        self.assertEqual(def_type, "interstitial")
        self.assertDictEqual(comp_diff, {"F": 1})
        (
            bulk_site_idx,
            defect_site_idx,
            unrelaxed_defect_structure,
        ) = parse_calculations.get_defect_site_idxs_and_unrelaxed_structure(
            bulk_sc_structure, initial_defect_structure, def_type, comp_diff
        )
        self.assertEqual(bulk_site_idx, None)
        self.assertEqual(defect_site_idx, len(unrelaxed_defect_structure) - 1)

        # assert auto-determined interstitial site is correct
        print(unrelaxed_defect_structure[defect_site_idx].frac_coords)
        self.assertAlmostEqual(
            unrelaxed_defect_structure[
                defect_site_idx
            ].distance_and_image_from_frac_coords(
                [-0.0005726049122470, -0.0001544430438804, 0.47800736578014720]
            )[
                0
            ],
            0.0,
            places=2,
        )  # approx match, not exact because relaxed bulk supercell

    def test_extrinsic_substitution_defect_ID(self):
        """Test parsing of extrinsic U_on_Cd in CdTe"""
        bulk_sc_structure = Structure.from_file(
            f"{self.CDTE_EXAMPLE_DIR}/CdTe_bulk_supercell_POSCAR"
        )
        initial_defect_structure = Structure.from_file(
            f"{self.CDTE_EXAMPLE_DIR}/U_on_Cd_POSCAR"
        )
        (
            def_type,
            comp_diff,
        ) = parse_calculations.get_defect_type_and_composition_diff(
            bulk_sc_structure, initial_defect_structure
        )
        self.assertEqual(def_type, "substitution")
        self.assertDictEqual(comp_diff, {"Cd": -1, "U": 1})
        (
            bulk_site_idx,
            defect_site_idx,
            unrelaxed_defect_structure,
        ) = parse_calculations.get_defect_site_idxs_and_unrelaxed_structure(
            bulk_sc_structure, initial_defect_structure, def_type, comp_diff
        )
        self.assertEqual(bulk_site_idx, 0)
        self.assertEqual(defect_site_idx, 63)  # last site in structure

        # assert auto-determined substitution site is correct
        np.testing.assert_array_almost_equal(
            unrelaxed_defect_structure[defect_site_idx].frac_coords,
            [0.00, 0.00, 0.00],
            decimal=2,  # exact match because perfect supercell
        )

    def test_extrinsic_interstitial_parsing_and_kumagai(self):
        """Test parsing of extrinsic F in YTOS interstitial and Kumagai-Oba (eFNV) correction"""

        for i in os.listdir(self.EXAMPLE_DIR):
            if "YTOS" in i:
                defect_file_path = f"{self.EXAMPLE_DIR}/{i}/Int_F_-1/"
                defect_charge = -1
                # parse with no transformation.json:
                sdp = parse_calculations.SingleDefectParser.from_paths(
                    path_to_defect=defect_file_path,
                    path_to_bulk=f"{self.EXAMPLE_DIR}/{i}/Bulk/",
                    dielectric=self.ytos_dielectric,
                    defect_charge=defect_charge,
                )
                sdp.kumagai_loader()
                sdp.get_stdrd_metadata()
                sdp.get_bulk_gap_data()
                sdp.run_compatibility()
                int_F_minus1_ent = sdp.defect_entry

        self.assertAlmostEqual(int_F_minus1_ent.energy, 0.767, places=3)
        self.assertAlmostEqual(int_F_minus1_ent.uncorrected_energy, 0.7515, places=3)
        correction_dict = {
            "charge_correction": 0.0155169495708003,
            "bandfilling_correction": -0.0,
            "bandedgeshifting_correction": 0.0,
        }
        for correction_name, correction_energy in correction_dict.items():
            self.assertAlmostEqual(
                int_F_minus1_ent.corrections[correction_name],
                correction_energy,
                places=3,
            )

        # assert auto-determined interstitial site is correct
        self.assertAlmostEqual(
            int_F_minus1_ent.site.distance_and_image_from_frac_coords([0, 0, 0.4847])[
                0
            ],
            0.0,
            places=2,
        )  # approx match, not exact because relaxed bulk supercell

        os.remove("bulk_voronoi_nodes.json")

    def test_extrinsic_substitution_parsing_and_freysoldt_and_kumagai(self):
        """
        Test parsing of extrinsic F-on-O substitution in YTOS,
        w/Kumagai-Oba (eFNV) and Freysoldt (FNV) corrections
        """

        # first using Freysoldt (FNV) correction
        for i in os.listdir(self.EXAMPLE_DIR):
            if "YTOS" in i:
                defect_file_path = f"{self.EXAMPLE_DIR}/{i}/F_O_1/"
                defect_charge = +1
                # parse with no transformation.json:
                sdp = parse_calculations.SingleDefectParser.from_paths(
                    path_to_defect=defect_file_path,
                    path_to_bulk=f"{self.EXAMPLE_DIR}/{i}/Bulk/",
                    dielectric=self.ytos_dielectric,
                    defect_charge=defect_charge,
                )
                sdp.freysoldt_loader()
                sdp.get_stdrd_metadata()
                sdp.get_bulk_gap_data()
                sdp.run_compatibility()
                F_O_1_ent = sdp.defect_entry

        self.assertAlmostEqual(F_O_1_ent.energy, 0.0308, places=3)
        self.assertAlmostEqual(F_O_1_ent.uncorrected_energy, -0.0852, places=3)
        correction_dict = {
            "charge_correction": 0.11602321245859282,
            "bandfilling_correction": -0.0,
            "bandedgeshifting_correction": 0.0,
        }
        for correction_name, correction_energy in correction_dict.items():
            self.assertAlmostEqual(
                F_O_1_ent.corrections[correction_name], correction_energy, places=3
            )
        # assert auto-determined interstitial site is correct
        self.assertAlmostEqual(
            F_O_1_ent.site.distance_and_image_from_frac_coords([0, 0, 0])[0],
            0.0,
            places=2,
        )

        # now using Kumagai-Oba (eFNV) correction
        for i in os.listdir(self.EXAMPLE_DIR):
            if "YTOS" in i:
                defect_file_path = f"{self.EXAMPLE_DIR}/{i}/F_O_1/"
                defect_charge = +1
                # parse with no transformation.json:
                sdp = parse_calculations.SingleDefectParser.from_paths(
                    path_to_defect=defect_file_path,
                    path_to_bulk=f"{self.EXAMPLE_DIR}/{i}/Bulk/",
                    dielectric=self.ytos_dielectric,
                    defect_charge=defect_charge,
                )
                sdp.kumagai_loader()
                sdp.get_stdrd_metadata()
                sdp.get_bulk_gap_data()
                sdp.run_compatibility()
                F_O_1_ent = sdp.defect_entry

        self.assertAlmostEqual(F_O_1_ent.energy, -0.0031, places=3)
        self.assertAlmostEqual(F_O_1_ent.uncorrected_energy, -0.0852, places=3)
        correction_dict = {
            "charge_correction": 0.08214,
            "bandfilling_correction": -0.0,
            "bandedgeshifting_correction": 0.0,
        }
        for correction_name, correction_energy in correction_dict.items():
            self.assertAlmostEqual(
                F_O_1_ent.corrections[correction_name], correction_energy, places=3
            )
        # assert auto-determined interstitial site is correct
        self.assertAlmostEqual(
            F_O_1_ent.site.distance_and_image_from_frac_coords([0, 0, 0])[0],
            0.0,
            places=2,
        )

    def test_voronoi_structure_mismatch_and_reparse(self):
        """
        Test that a mismatch in bulk_supercell structure from previously parsed
        Voronoi nodes json file with current defect bulk supercell is detected and
        re-parsed
        """
        with patch("builtins.print") as mock_print:
            for i in os.listdir(self.CDTE_EXAMPLE_DIR):
                if "Int_Te" in i:  # loop folders and parse those with "Int_Te" in name
                    defect_file_path = f"{self.CDTE_EXAMPLE_DIR}/{i}/vasp_ncl"
                    defect_charge = int(i[-2:].replace("_", ""))
                    # parse with no transformation.json:
                    sdp = parse_calculations.SingleDefectParser.from_paths(
                        path_to_defect=defect_file_path,
                        path_to_bulk=self.CDTE_BULK_DATA_DIR,
                        dielectric=self.cdte_dielectric,
                        defect_charge=defect_charge,
                    )

        mock_print.assert_called_once_with(
            "Saving parsed Voronoi sites (for interstitial site-matching) "
            "to bulk_voronoi_sites.json to speed up future parsing."
        )

        with warnings.catch_warnings(record=True) as w:
            for i in os.listdir(self.EXAMPLE_DIR):
                if "YTOS" in i:
                    defect_file_path = f"{self.EXAMPLE_DIR}/{i}/Int_F_-1/"
                    defect_charge = -1
                    # parse with no transformation.json:
                    sdp = parse_calculations.SingleDefectParser.from_paths(
                        path_to_defect=defect_file_path,
                        path_to_bulk=f"{self.EXAMPLE_DIR}/{i}/Bulk/",
                        dielectric=self.ytos_dielectric,
                        defect_charge=defect_charge,
                    )

        warning_message = (
            "Previous bulk_voronoi_nodes.json detected, but does not "
            "match current bulk supercell. Recalculating Voronoi nodes."
        )
        user_warnings = [warning for warning in w if warning.category == UserWarning]
        self.assertEqual(len(user_warnings), 1)
        self.assertIn(warning_message, str(user_warnings[0].message))
        os.remove("bulk_voronoi_nodes.json")


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import unittest

from monty.serialization import loadfn
from pymatgen.io.vasp.inputs import Incar, Kpoints, Poscar

from doped import vasp_input


class VaspInputTestCase(unittest.TestCase):
    def setUp(self):
        self.DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
        self.CDTE_DATA_DIR = os.path.join(self.DATA_DIR, "CdTe")
        self.cdte_generated_defect_dict = loadfn(
            os.path.join(self.CDTE_DATA_DIR, "CdTe_generated_defect_dict.json")
        )

    def tearDown(self):
        for folder in os.listdir(
            self.CDTE_DATA_DIR
        ):  # remove all generated CdTe defect folders
            if os.path.isdir(folder):
                shutil.rmtree(folder)

    def check_generated_vasp_inputs(
        self,
        data_dir=None,
        generated_dir=".",
        vasp_type="vasp_gam",
        check_poscar=True,
        check_potcar_spec=False,
        single_defect_dir=False,
    ):
        def _check_single_vasp_dir(
            data_dir=None,
            generated_dir=".",
            folder="",
            vasp_type="vasp_gam",
            check_poscar=True,
            check_potcar_spec=False,
        ):
            self.assertTrue(os.path.exists(f"{generated_dir}/{folder}"))
            self.assertTrue(os.path.exists(f"{generated_dir}/{folder}/{vasp_type}"))

            # load the Incar, Poscar and Kpoints and check it matches the previous:
            test_incar = Incar.from_file(f"{data_dir}/{folder}/{vasp_type}/INCAR")
            incar = Incar.from_file(f"{generated_dir}/{folder}/{vasp_type}/INCAR")
            # remove NUPDOWN and NELECT entries from test_incar:
            test_incar.pop("NUPDOWN", None)
            incar.pop("NUPDOWN", None)
            test_incar.pop("NELECT", None)
            incar.pop("NELECT", None)
            self.assertEqual(test_incar, incar)

            if check_poscar:
                test_poscar = Poscar.from_file(
                    f"{data_dir}/{folder}/vasp_gam/POSCAR"  # POSCAR always checked
                    # against vasp_gam unperturbed POSCAR
                )
                poscar = Poscar.from_file(
                    f"{generated_dir}/{folder}/{vasp_type}/POSCAR"
                )
                self.assertEqual(test_poscar.structure, poscar.structure)

            if check_potcar_spec:
                with open(
                    f"{generated_dir}/{folder}/{vasp_type}/POTCAR.spec", "r"
                ) as file:
                    contents = file.readlines()
                    self.assertIn(contents[0], ["Cd", "Cd\n"])
                    self.assertIn(contents[1], ["Te", "Te\n"])
                    if "Se" in folder:
                        self.assertIn(contents[2], ["Se", "Se\n"])

            test_kpoints = Kpoints.from_file(f"{data_dir}/{folder}/{vasp_type}/KPOINTS")
            kpoints = Kpoints.from_file(f"{generated_dir}/{folder}/{vasp_type}/KPOINTS")
            self.assertDictEqual(test_kpoints.as_dict(), kpoints.as_dict())

        if data_dir is None:
            data_dir = self.CDTE_DATA_DIR

        if single_defect_dir:
            _check_single_vasp_dir(
                data_dir=data_dir,
                generated_dir=generated_dir,
                folder="",
                vasp_type=vasp_type,
                check_poscar=check_poscar,
                check_potcar_spec=check_potcar_spec,
            )

        else:
            for folder in os.listdir(data_dir):
                if os.path.isdir(f"{data_dir}/{folder}"):
                    _check_single_vasp_dir(
                        data_dir=data_dir,
                        generated_dir=generated_dir,
                        folder=folder,
                        vasp_type=vasp_type,
                        check_poscar=check_poscar,
                        check_potcar_spec=check_potcar_spec,
                    )

    def test_vasp_gam_files(self):
        vasp_input.vasp_gam_files(
            self.cdte_generated_defect_dict,
            user_incar_settings={"ENCUT": 350},
            potcar_spec=True,
            user_potcar_functional=None,
        )

        # assert that the same folders in self.CDTE_DATA_DIR are present in the current directory
        self.check_generated_vasp_inputs(check_potcar_spec=True)

        # test custom POTCAR choice:
        vac_Te_dict = {
            defect_type: [
                defect for defect in defect_list if defect["name"] == "vac_2_Te"
            ]
            for defect_type, defect_list in self.cdte_generated_defect_dict.items()
            if defect_type == "vacancies"
        }

        vasp_input.vasp_gam_files(
            vac_Te_dict,
            user_potcar_settings={"Cd": "Cd_sv_GW", "Te": "Te_sv_GW"},
            user_potcar_functional=None,
            potcar_spec=True,
        )

        with open("vac_2_Te_0/vasp_gam/POTCAR.spec", "r") as file:
            contents = file.readlines()
            self.assertIn(contents[0], ["Cd_sv_GW", "Cd_sv_GW\n"])
            self.assertIn(contents[1], ["Te_sv_GW", "Te_sv_GW\n"])

        # test no files written with write_files=False
        self.tearDown()
        vasp_input.vasp_gam_files(
            self.cdte_generated_defect_dict,
            user_potcar_functional=None,
            write_files=False
        )
        for folder in os.listdir(self.CDTE_DATA_DIR):
            if os.path.isdir(f"{self.CDTE_DATA_DIR}/{folder}"):
                self.assertFalse(os.path.exists(f"./{folder}"))


    def test_vasp_gam_files_single_defect_dict(self):
        single_defect_dict = self.cdte_generated_defect_dict["vacancies"][0]  # V_Cd
        vasp_input.vasp_gam_files(
            single_defect_dict,
            user_incar_settings={"ENCUT": 350},
            potcar_spec=True,
            user_potcar_functional=None,
        )

        # assert that the same folders in self.CDTE_DATA_DIR are present in the current directory
        for charge in range(-2, 1):
            self.check_generated_vasp_inputs(
                generated_dir=f"vac_1_Cd_{charge}",
                data_dir=f"{self.CDTE_DATA_DIR}/vac_1_Cd_{charge}",
                check_potcar_spec=True,
                single_defect_dir=True,
            )

        # test custom POTCAR choice:
        vac_Te_dict = self.cdte_generated_defect_dict["vacancies"][1]  # V_Te

        vasp_input.vasp_gam_files(
            vac_Te_dict,
            user_potcar_settings={"Cd": "Cd_sv_GW", "Te": "Te_sv_GW"},
            user_potcar_functional=None,
            potcar_spec=True,
        )

        with open("vac_2_Te_0/vasp_gam/POTCAR.spec", "r") as file:
            contents = file.readlines()
            self.assertIn(contents[0], ["Cd_sv_GW", "Cd_sv_GW\n"])
            self.assertIn(contents[1], ["Te_sv_GW", "Te_sv_GW\n"])

        # test no files written with write_files=False
        self.tearDown()
        vasp_input.vasp_gam_files(
            vac_Te_dict,
            user_potcar_functional=None,
            write_files=False
        )
        self.assertFalse(os.path.exists(f"vac_2_Te_0"))

    def test_vasp_std_files(self):
        vasp_input.vasp_std_files(
            self.cdte_generated_defect_dict,
            user_incar_settings={
                "ENCUT": 350,
                "LREAL": "Auto",
                "IBRION": 5,
                "ADDGRID": True,
            },
            user_potcar_functional=None,
        )

        # assert that the same folders in self.CDTE_DATA_DIR are present in the current directory
        self.check_generated_vasp_inputs(vasp_type="vasp_std", check_poscar=False)

        vasp_input.vasp_std_files(
            self.cdte_generated_defect_dict,
            user_incar_settings={
                "ENCUT": 350,
                "LREAL": "Auto",
                "IBRION": 5,
                "ADDGRID": True,
            },
            unperturbed_poscar=True,
            user_potcar_functional=None,
        )
        self.check_generated_vasp_inputs(vasp_type="vasp_std", check_poscar=True)

        # test no files written with write_files=False
        self.tearDown()
        vasp_input.vasp_std_files(
            self.cdte_generated_defect_dict,
            user_potcar_functional=None,
            write_files=False
        )
        for folder in os.listdir(self.CDTE_DATA_DIR):
            if os.path.isdir(f"{self.CDTE_DATA_DIR}/{folder}"):
                self.assertFalse(os.path.exists(f"./{folder}"))

    def test_vasp_std_files_single_defect_dict(self):
        # test interstitials this time:
        single_defect_dict = self.cdte_generated_defect_dict["interstitials"][
            0
        ]  # inter_1_Cd
        vasp_input.vasp_std_files(
            single_defect_dict,
            user_incar_settings={
                "ENCUT": 350,
                "LREAL": "Auto",
                "IBRION": 5,
                "ADDGRID": True,
            },
            user_potcar_functional=None,
        )

        # assert that the same folders in self.CDTE_DATA_DIR are present in the current directory
        for charge in range(0, 3):
            self.check_generated_vasp_inputs(
                generated_dir=f"inter_1_Cd_{charge}",
                data_dir=f"{self.CDTE_DATA_DIR}/inter_1_Cd_{charge}",
                vasp_type="vasp_std",
                check_poscar=False,
                single_defect_dir=True,
            )

        # test unperturbed POSCAR:
        vasp_input.vasp_std_files(
            single_defect_dict,
            user_incar_settings={
                "ENCUT": 350,
                "LREAL": "Auto",
                "IBRION": 5,
                "ADDGRID": True,
            },
            user_potcar_functional=None,
            unperturbed_poscar=True,
        )

        for charge in range(0, 3):
            self.check_generated_vasp_inputs(
                generated_dir=f"inter_1_Cd_{charge}",
                data_dir=f"{self.CDTE_DATA_DIR}/inter_1_Cd_{charge}",
                vasp_type="vasp_std",
                check_poscar=True,
                single_defect_dir=True,
            )

        # test no files written with write_files=False
        self.tearDown()
        vasp_input.vasp_gam_files(
            single_defect_dict,
            user_potcar_functional=None,
            write_files=False
        )
        self.assertFalse(os.path.exists(f"inter_1_Cd_0"))
        self.assertFalse(os.path.exists(f"inter_1_Cd_1"))
        self.assertFalse(os.path.exists(f"inter_1_Cd_2"))

    def test_vasp_ncl_files(self):
        vasp_input.vasp_ncl_files(
            self.cdte_generated_defect_dict,
            user_incar_settings={"ENCUT": 750, "LREAL": True, "ADDGRID": False},
            user_potcar_functional=None,
        )

        # assert that the same folders in self.CDTE_DATA_DIR are present in the current directory
        self.check_generated_vasp_inputs(vasp_type="vasp_ncl", check_poscar=False)

        vasp_input.vasp_ncl_files(
            self.cdte_generated_defect_dict,
            user_incar_settings={"ENCUT": 750, "LREAL": True, "ADDGRID": False},
            unperturbed_poscar=True,
            user_potcar_functional=None,
        )
        self.check_generated_vasp_inputs(vasp_type="vasp_ncl", check_poscar=True)

        # test no files written with write_files=False
        self.tearDown()
        vasp_input.vasp_ncl_files(
            self.cdte_generated_defect_dict,
            user_potcar_functional=None,
            write_files=False
        )
        for folder in os.listdir(self.CDTE_DATA_DIR):
            if os.path.isdir(f"{self.CDTE_DATA_DIR}/{folder}"):
                self.assertFalse(os.path.exists(f"./{folder}"))

    def test_vasp_ncl_files_single_defect_dict(self):
        # test substitutions this time:
        single_defect_dict = self.cdte_generated_defect_dict["substitutions"][
            0
        ]  # sub_2_Se_on_Te
        vasp_input.vasp_ncl_files(
            single_defect_dict,
            user_incar_settings={"ENCUT": 750, "LREAL": True, "ADDGRID": False},
            user_potcar_functional=None,
        )

        # assert that the same folders in self.CDTE_DATA_DIR are present in the current directory
        for charge in range(-1, 6):
            self.check_generated_vasp_inputs(
                generated_dir=f"sub_2_Se_on_Te_{charge}",
                data_dir=f"{self.CDTE_DATA_DIR}/sub_2_Se_on_Te_{charge}",
                vasp_type="vasp_ncl",
                check_poscar=False,
                single_defect_dir=True,
            )

        # test unperturbed POSCAR:
        vasp_input.vasp_ncl_files(
            single_defect_dict,
            user_incar_settings={"ENCUT": 750, "LREAL": True, "ADDGRID": False},
            user_potcar_functional=None,
            unperturbed_poscar=True,
        )

        for charge in range(-1, 6):
            self.check_generated_vasp_inputs(
                generated_dir=f"sub_2_Se_on_Te_{charge}",
                data_dir=f"{self.CDTE_DATA_DIR}/sub_2_Se_on_Te_{charge}",
                vasp_type="vasp_ncl",
                check_poscar=True,
                single_defect_dir=True,
            )

        # test no files written with write_files=False
        self.tearDown()
        vasp_input.vasp_ncl_files(
            single_defect_dict,
            user_potcar_functional=None,
            write_files=False
        )
        for charge in range(-1, 6):
            self.assertFalse(os.path.exists(f"sub_2_Se_on_Te_{charge}"))


if __name__ == "__main__":
    unittest.main()



# coding: utf-8

"""
Code to analyse VASP defect calculations.
These functions are built from a combination of useful modules from pymatgen and AIDE (by Adam
Jackson and Alex Ganose), alongside substantial modification, in the efforts of making an
efficient, user-friendly package for managing and analysing defect calculations,
with publication-quality outputs
"""

import warnings

from matplotlib import cm, ticker, rc
import matplotlib.pyplot as plt
import numpy as np
from pymatgen.util.string import latexify

default_fonts = [
    "Whitney Book Extended",
    "Whitney Pro",
    "Arial",
    "Whitney Book",
    "Helvetica",
    "Liberation Sans",
    "Andale Sans",
]


# TODO: Lean out the options for this function (inherited from AIDE)(particularly those that can
#  just be edited by the user with the returned Matplotlib object – show example of this in
#  notebooks maybe?)
def formation_energy_plot(
    defect_phase_diagram,
    chempot_limits: dict = None,
    elt_refs: dict = None,
    ax=None,
    fonts=None,
    xlim=None,
    ylim=None,
    ax_fontsize=1.0,
    lg_fontsize=1.0,
    lg_position=None,
    fermi_level=None,
    title: str = None,
    saved=False,
    colormap="Dark2",
    frameon=False,
    chempot_table=True,
    pd_facets: list = None,
    auto_labels: bool = False,
    filename: str = None,
    emphasis=False,
):
    """
    Produce a defect formation energy vs Fermi energy plot (i.e. a defect transition level diagram).

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
             DefectPhaseDiagram object (likely created from analysis.dpd_from_parsed_defect_dict)
        chempot_limits (dict):
            This can either be a dictionary of chosen absolute/DFT chemical potentials: {Elt:
            Energy} (giving a single formation energy table – recommended to use the elt_refs
            option with this to show the formal (relative) chemical potentials in the plot) or a
            dictionary including the key-value pair: {"facets": [{'facet': [chempot_dict]}]},
            following the format generated by doped: cpa.read_phase_diagram_and_chempots() (see
            example notebooks). If not specified, chemical potentials are not included in the
            formation energy calculation (all set to zero energy).
        elt_refs (dict):
            Dictionary of elemental reference energies for the chemical potentials in the format:
            {Elt: ref_energy} (to determine the formal chemical potentials, when chempot_limits
            is specified as a dictionary {Elt: Energy}). Unnecessary if chempot_limits is
            provided in format generated by doped: cpa.read_phase_diagram_and_chempots() (see
            example notebooks).
            Default: None
        pd_facets (list):
            A list facet(s) / chemical potential limit(s) for which to print the defect formation
            energy tables. If not specified, will print formation energy tables for each facet in
            the phase diagram. (default: None)
        fermi_level (float):
            Fermi level to use for computing the defect formation energies. (default: 0 (i.e.
            at the VBM))
        xlim:
            Tuple (min,max) giving the range of the x (fermi energy) axis. This may need to be
            set manually when including transition level labels, so that they don't cross the axes.
        ylim:
            Tuple (min,max) giving the range for the formation energy axis. This may need to be
            set manually when including transition level labels, so that they don't cross the axes.
        ax_fontsize:
            float  multiplier to change axis label fontsize
        lg_fontsize:
            float  multiplier to change legend label fontsize
        lg_position:
            Tuple (horizontal-position, vertical-position) giving the position
            to place the legend.
            Example: (0.5,-0.75) will likely put it below the x-axis.
    Returns:
        a matplotlib object
    """
    if chempot_limits and "facets" in chempot_limits:
        if not pd_facets:
            pd_facets = chempot_limits[
                "facets"
            ].keys()  # Phase diagram facets to use for chemical
            # potentials, to calculate and plot formation energies
        for facet in pd_facets:
            mu_elts = chempot_limits["facets"][facet]
            elt_refs = chempot_limits["elemental_refs"]
            if title:
                plot_title = title
            else:
                plot_title = facet
            if filename:
                plot_filename = filename
            else:
                plot_filename = plot_title + "_" + facet + ".pdf"

            plot = _aide_pmg_plot(
                defect_phase_diagram,
                mu_elts=mu_elts,
                elt_refs=elt_refs,
                ax=ax,
                fonts=fonts,
                xlim=xlim,
                ylim=ylim,
                ax_fontsize=ax_fontsize,
                lg_fontsize=lg_fontsize,
                lg_position=lg_position,
                fermi_level=fermi_level,
                title=plot_title,
                saved=saved,
                colormap=colormap,
                frameon=frameon,
                chempot_table=chempot_table,
                auto_labels=auto_labels,
                filename=plot_filename,
                emphasis=emphasis,
            )

        return plot

    # Else if you only want to give {Elt: Energy} dict for chempot_limits, or no chempot_limits
    return _aide_pmg_plot(
        defect_phase_diagram,
        mu_elts=chempot_limits,
        elt_refs=elt_refs,
        ax=ax,
        fonts=fonts,
        xlim=xlim,
        ylim=ylim,
        ax_fontsize=ax_fontsize,
        lg_fontsize=lg_fontsize,
        lg_position=lg_position,
        fermi_level=fermi_level,
        title=title,
        saved=saved,
        colormap=colormap,
        frameon=frameon,
        chempot_table=chempot_table,
        auto_labels=auto_labels,
        filename=filename,
        emphasis=emphasis,
    )


def _aide_pmg_plot(
    defect_phase_diagram,
    mu_elts=None,
    elt_refs=None,
    ax=None,
    fonts=None,
    xlim=None,
    ylim=None,
    ax_fontsize=1.0,
    lg_fontsize=1.0,
    lg_position=None,
    fermi_level=None,
    title=None,
    saved=False,
    colormap="Dark2",
    frameon=False,
    chempot_table=True,
    auto_labels=False,
    filename=None,
    emphasis=False,
):
    """
    Produce defect Formation energy vs Fermi energy plot
    Args:
        mu_elts:
            a dictionary of {Element:value} giving the chemical
            potential of each element
        xlim:
            Tuple (min,max) giving the range of the x (fermi energy) axis. This may need to be
            set manually when including transition level labels, so that they dont' cross the axes.
        ylim:
            Tuple (min,max) giving the range for the formation energy axis. This may need to be
            set manually when including transition level labels, so that they dont' cross the axes.
        ax_fontsize:
            float  multiplier to change axis label fontsize
        lg_fontsize:
            float  multiplier to change legend label fontsize
        lg_position:
            Tuple (horizontal-position, vertical-position) giving the position
            to place the legend.
            Example: (0.5,-0.75) will likely put it below the x-axis.
    Returns:
        a matplotlib object
    """
    if mu_elts is None:
        warnings.warn(
            "No chemical potentials specified, so chemical potentials are set to zero "
            "for each species. Note that this will give large errors in the absolute "
            "values of formation energies, but the transition level positions will be "
            "unaffected."
        )

    if xlim is None:
        xlim = (-0.4, defect_phase_diagram.band_gap + 0.4)
    xy = {}
    all_lines_xy = (
        {}
    )  # For emphasis plots with faded grey E_form lines for all charge states
    lower_cap = -100.0
    upper_cap = 100.0
    y_range_vals = []  # for finding max/min values on y-axis based on x-limits

    for defnom, def_tl in defect_phase_diagram.transition_level_map.items():
        xy[defnom] = [[], []]
        if emphasis:
            all_lines_xy[defnom] = [[], []]
            for chg_ent in defect_phase_diagram.stable_entries[defnom]:
                for x_extrem in [lower_cap, upper_cap]:
                    all_lines_xy[defnom][0].append(x_extrem)
                    all_lines_xy[defnom][1].append(
                        chg_ent.formation_energy(
                            chemical_potentials=mu_elts, fermi_level=x_extrem
                        )
                    )
                # for x_window in xlim:
                #    y_range_vals.append(
                #        chg_ent.formation_energy(chemical_potentials=mu_elts, fermi_level=x_window)
                #    )

        if def_tl:
            org_x = list(def_tl.keys())  # list of transition levels
            org_x.sort()  # sorted with lowest first

            # establish lower x-bound
            first_charge = max(def_tl[org_x[0]])
            for chg_ent in defect_phase_diagram.stable_entries[defnom]:
                if chg_ent.charge == first_charge:
                    form_en = chg_ent.formation_energy(
                        chemical_potentials=mu_elts, fermi_level=lower_cap
                    )
                    fe_left = chg_ent.formation_energy(
                        chemical_potentials=mu_elts, fermi_level=xlim[0]
                    )
            xy[defnom][0].append(lower_cap)
            xy[defnom][1].append(form_en)
            y_range_vals.append(fe_left)
            # iterate over stable charge state transitions
            for fl in org_x:
                charge = max(def_tl[fl])
                for chg_ent in defect_phase_diagram.stable_entries[defnom]:
                    if chg_ent.charge == charge:
                        form_en = chg_ent.formation_energy(
                            chemical_potentials=mu_elts, fermi_level=fl
                        )
                xy[defnom][0].append(fl)
                xy[defnom][1].append(form_en)
                y_range_vals.append(form_en)
            # establish upper x-bound
            last_charge = min(def_tl[org_x[-1]])
            for chg_ent in defect_phase_diagram.stable_entries[defnom]:
                if chg_ent.charge == last_charge:
                    form_en = chg_ent.formation_energy(
                        chemical_potentials=mu_elts, fermi_level=upper_cap
                    )
                    fe_right = chg_ent.formation_energy(
                        chemical_potentials=mu_elts, fermi_level=xlim[1]
                    )
            xy[defnom][0].append(upper_cap)
            xy[defnom][1].append(form_en)
            y_range_vals.append(fe_right)
        else:
            # no transition - just one stable charge
            chg_ent = defect_phase_diagram.stable_entries[defnom][0]
            for x_extrem in [lower_cap, upper_cap]:
                xy[defnom][0].append(x_extrem)
                xy[defnom][1].append(
                    chg_ent.formation_energy(
                        chemical_potentials=mu_elts, fermi_level=x_extrem
                    )
                )
            for x_window in xlim:
                y_range_vals.append(
                    chg_ent.formation_energy(
                        chemical_potentials=mu_elts, fermi_level=x_window
                    )
                )

    cmap = cm.get_cmap(colormap)
    colors = cmap(np.linspace(0, 1, len(xy)))
    if colormap == "Dark2" and len(xy) >= 8:
        warnings.warn(
            f"""
The chosen colormap is Dark2, which only has 8 colours, yet you have {len(xy)} defect species (so
some defects will have the same line colour). Recommended to change/set colormap to 'tab10' or
'tab20' (10 and 20 colours each)."""
        )
    plt.figure(dpi=600, figsize=(2.6, 1.95))  # Gives a final figure width of c. 3.5
    # inches, the standard single column width for publication (which is what we're about)
    plt.clf()
    width = 9
    ax = pretty_axis(ax=ax, fonts=fonts)
    # plot formation energy lines
    for_legend = []
    for cnt, defnom in enumerate(xy.keys()):
        ax.plot(
            xy[defnom][0],
            xy[defnom][1],
            color=colors[cnt],
            markeredgecolor=colors[cnt],
            lw=1.2,
            markersize=3.5,
        )
        for_legend.append(defect_phase_diagram.stable_entries[defnom][0].copy())
    # Redo for loop so grey 'all_lines_xy' not included in legend
    for cnt, defnom in enumerate(xy.keys()):
        if emphasis:
            ax.plot(
                all_lines_xy[defnom][0],
                all_lines_xy[defnom][1],
                color=(0.8, 0.8, 0.8),
                markeredgecolor=colors[cnt],
                lw=1.2,
                markersize=3.5,
                alpha=0.5,
            )
    # plot transition levels
    for cnt, defnom in enumerate(xy.keys()):
        x_trans, y_trans = [], []
        tl_labels = []
        tl_label_type = []
        for x_val, chargeset in defect_phase_diagram.transition_level_map[
            defnom
        ].items():
            x_trans.append(x_val)
            for chg_ent in defect_phase_diagram.stable_entries[defnom]:
                if chg_ent.charge == chargeset[0]:
                    form_en = chg_ent.formation_energy(
                        chemical_potentials=mu_elts, fermi_level=x_val
                    )
            y_trans.append(form_en)
            tl_labels.append(
                rf"$\epsilon$({max(chargeset):{'+' if max(chargeset) else ''}}/"
                f"{min(chargeset):{'+' if min(chargeset) else ''}})"
            )
            tl_label_type.append(
                "start_positive" if max(chargeset) > 0 else "end_negative"
            )
        if x_trans:
            ax.plot(
                x_trans,
                y_trans,
                marker="o",
                color=colors[cnt],
                markeredgecolor=colors[cnt],
                lw=1.2,
                markersize=3.5,
                fillstyle="full",
            )
            if auto_labels:
                for index, coords in enumerate(zip(x_trans, y_trans)):
                    text_alignment = (
                        "right" if tl_label_type[index] == "start_positive" else "left"
                    )
                    ax.annotate(
                        tl_labels[index],  # this is the text
                        coords,  # this is the point to label
                        textcoords="offset points",  # how to position the text
                        xytext=(0, 5),  # distance from text to points (x,y)
                        ha=text_alignment,  # horizontal alignment of text
                        size=ax_fontsize * width * 0.9,
                        annotation_clip=True,
                    )  # only show label if coords in current axes

    # get latex-like legend titles
    legends_txt = []
    for dfct in for_legend:
        flds = dfct.name.split("_")
        if flds[0] == "Vac":
            base = r"$\mathrm{V"
            sub_str = "_{" + flds[1] + "}}$"
        elif flds[0] == "Sub":
            flds = dfct.name.split("_")
            base = r"$\mathrm{" + flds[1]
            sub_str = "_{" + flds[3] + "}}$"
        elif flds[0] == "Int":
            base = r"$\mathrm{" + flds[1]
            sub_str = "_{i}}$"
        else:
            base = dfct.name
            sub_str = ""
        def_name = base + sub_str
        # add subscript labels for different configurations of same defect species
        labelled_def_name = def_name + r"$_{, 1}$"
        if def_name in legends_txt:
            def_name = labelled_def_name
        if def_name in legends_txt:
            i = 1
            while def_name in legends_txt:
                i += 1
                def_name = def_name[:-3] + f"{i}" + def_name[-2:]
            legends_txt.append(def_name)
        else:
            legends_txt.append(def_name)

    if not lg_position:
        ax.legend(
            legends_txt,
            fontsize=lg_fontsize * width,
            loc=2,
            bbox_to_anchor=(1, 1),
            frameon=frameon,
            prop=fonts,
        )
    else:
        ax.legend(
            legends_txt,
            fontsize=lg_fontsize * width,
            ncol=3,
            loc="lower center",
            bbox_to_anchor=lg_position,
        )

    if ylim is None:
        window = max(y_range_vals) - min(y_range_vals)
        spacer = 0.1 * window
        ylim = (0, max(y_range_vals) + spacer)
        if auto_labels:  # need to manually set xlim or ylim if labels cross axes!!
            ylim = (0, max(y_range_vals) * 1.17) if spacer / ylim[1] < 0.145 else ylim
            # Increase y_limit to give space for transition level labels

    # Show colourful band edges
    ax.imshow(
        [(0, 1), (0, 1)],
        cmap=plt.cm.Blues,
        extent=(xlim[0], 0, ylim[0], ylim[1]),
        vmin=0,
        vmax=3,
        interpolation="bicubic",
        rasterized=True,
        aspect="auto",
    )

    ax.imshow(
        [(1, 0), (1, 0)],
        cmap=plt.cm.Oranges,
        extent=(defect_phase_diagram.band_gap, xlim[1], ylim[0], ylim[1]),
        vmin=0,
        vmax=3,
        interpolation="bicubic",
        rasterized=True,
        aspect="auto",
    )

    ax.set_ylim(ylim)
    ax.set_xlim(xlim)
    # ax.plot([xlim[0], xlim[1]], [0, 0], "k-")  # black dashed line for E_formation = 0

    if fermi_level is not None:
        plt.axvline(
            x=fermi_level, linestyle="-.", color="k", linewidth=1
        )  # smaller dashed lines for gap edges
    ax.set_xlabel("Fermi Level (eV)", size=ax_fontsize * width)
    ax.set_ylabel("Formation Energy (eV)", size=ax_fontsize * width)
    ax.xaxis.set_major_locator(ticker.MaxNLocator(4))
    ax.xaxis.set_minor_locator(ticker.AutoMinorLocator(2))
    ax.yaxis.set_major_locator(ticker.MaxNLocator(4))
    ax.yaxis.set_minor_locator(ticker.AutoMinorLocator(2))
    ax.xaxis.set_major_formatter(_CustomScalarFormatter(minus_symbol="-"))
    ax.yaxis.set_major_formatter(_CustomScalarFormatter(minus_symbol="-"))
    if chempot_table:
        if elt_refs is not None:
            facets_wrt_elt_refs = {
                elt: energy - elt_refs[elt] for elt, energy in mu_elts.items()
            }
            _plot_chemical_potential_table(
                plt,
                facets_wrt_elt_refs,
                fontsize=ax_fontsize * width,
                wrt_elt_refs=True,
            )
        elif mu_elts:
            _plot_chemical_potential_table(
                plt,
                mu_elts,
                fontsize=ax_fontsize * width,
                wrt_elt_refs=False,
            )

    if title and chempot_table:
        ax.set_title(
            latexify(title),
            size=1.2 * ax_fontsize * width,
            pad=28,
            fontdict={"fontweight": "bold"},
        )
    elif title:
        ax.set_title(
            latexify(title), size=ax_fontsize * width, fontdict={"fontweight": "bold"}
        )
    if saved or filename:
        if filename:
            plt.savefig(filename, bbox_inches="tight", dpi=600)
        else:
            plt.savefig(str(title) + "_doped_plot.pdf", bbox_inches="tight", dpi=600)
    return ax


def _plot_chemical_potential_table(
    plot,
    chemical_potentials,
    chempot_label="",
    fontsize=9,
    loc="left",
    ax=None,
    wrt_elt_refs=False,
):
    if ax is None:
        ax = plot.gca()

    labels = [""] + [
        rf"$\mathregular{{\mu_{{{s}}}}}$," for s in sorted(chemical_potentials.keys())
    ]
    # add if else here, to use 'facets' if no wrt_elts, and don't say wrt elt_refs etc.
    labels[1] = "(" + labels[1]
    labels[-1] = labels[-1][:-1] + ")"
    labels = ["Chemical Potentials"] + labels + [" Units:"]
    text = [[chempot_label]]

    for el in sorted(chemical_potentials.keys()):
        text[0].append(f"{chemical_potentials[el]:.2f},")

    text[0][1] = "(" + text[0][1]
    text[0][-1] = text[0][-1][:-1] + ")"
    if wrt_elt_refs:
        text[0] = ["(wrt Elemental refs)"] + text[0] + ["  [eV]"]
    else:
        text[0] = ["(from calculations)"] + text[0] + ["  [eV]"]
    widths = [0.1] + [0.9 / len(chemical_potentials)] * (len(chemical_potentials) + 2)
    tab = ax.table(
        cellText=text, colLabels=labels, colWidths=widths, loc="top", cellLoc=loc
    )
    tab.auto_set_font_size(False)
    tab.set_fontsize(fontsize)

    tab.auto_set_column_width(list(range(len(widths))))
    tab.scale(1.0, 1.0)  # Default spacing is based on fontsize, just bump it up
    for cell in tab.get_celld().values():
        cell.set_linewidth(0)

    return tab


class _CustomScalarFormatter(ticker.ScalarFormatter):
    """Derived matplotlib tick formatter for arbitrary minus signs

    Args:
        minus_symbol (str): Symbol used in place of hyphen"""

    def __init__(
        self,
        useOffset=None,
        useMathText=None,
        useLocale=None,
        minus_symbol="\N{MINUS SIGN}",
    ):
        self.minus_symbol = minus_symbol
        super().__init__(
            useOffset=useOffset, useMathText=useMathText, useLocale=useLocale
        )


def pretty_axis(ax=None, fonts=None):
    ticklabelsize = 9
    ticksize = 8
    linewidth = 1.0

    if ax is None:
        ax = plt.gca()

    ax.tick_params(width=linewidth, size=ticksize)
    ax.tick_params(
        which="major", size=ticksize, width=linewidth, labelsize=ticklabelsize, pad=3
    )
    ax.tick_params(which="minor", size=ticksize / 2, width=linewidth)

    ax.set_title(ax.get_title(), size=9.5)
    for axis in ["top", "bottom", "left", "right"]:
        ax.spines[axis].set_linewidth(linewidth)

    labelsize = int(9)

    ax.set_xlabel(ax.get_xlabel(), size=labelsize)
    ax.set_ylabel(ax.get_ylabel(), size=labelsize)

    fonts = default_fonts if fonts is None else fonts + default_fonts

    rc("font", **{"family": "sans-serif", "sans-serif": fonts})
    rc("text", usetex=False)
    rc("pdf", fonttype=42)
    # rc('mathtext', fontset='stixsans')

    return ax


def all_lines_formation_energy_plot(
    defect_phase_diagram,
    chempot_limits=None,
    elt_refs: dict = None,
    ax=None,
    fonts=None,
    xlim=None,
    ylim=None,
    ax_fontsize=1.0,
    lg_fontsize=1.0,
    lg_position=None,
    fermi_level=None,
    title=None,
    saved=False,
    colormap="Dark2",
    frameon=False,
    chempot_table=True,
    pd_facets: list = None,
    auto_labels: bool = False,
    filename: str = None,
):
    """
    Produce a defect formation energy vs Fermi energy plot (i.e. a defect transition level
    diagram), showing the full formation energy lines for all defect species present.

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
             DefectPhaseDiagram object (likely created from analysis.dpd_from_parsed_defect_dict)
        chempot_limits (dict):
            This can either be a dictionary of chosen absolute/DFT chemical potentials: {Elt:
            Energy} (giving a single formation energy table – recommended to use the elt_refs
            option with this to show the formal (relative) chemical potentials in the plot) or a
            dictionary including the key-value pair: {"facets": [{'facet': [chempot_dict]}]},
            following the format generated by doped: cpa.read_phase_diagram_and_chempots() (see
            example notebooks). If not specified, chemical potentials are not included in the
            formation energy calculation (all set to zero energy).
        elt_refs (dict):
            Dictionary of elemental reference energies for the chemical potentials in the format:
            {Elt: ref_energy} (to determine the formal chemical potentials, when chempot_limits
            is specified as a dictionary {Elt: Energy}). Unnecessary if chempot_limits is
            provided in format generated by doped: cpa.read_phase_diagram_and_chempots() (see
            example notebooks).
            Default: None
        pd_facets (list):
            A list facet(s) / chemical potential limit(s) for which to print the defect formation
            energy tables. If not specified, will print formation energy tables for each facet in
            the phase diagram.
            Default: None
        fermi_level (float):
            Fermi level to use for computing the defect formation energies. (default: 0 (i.e. at
            the VBM))
        xlim:
            Tuple (min,max) giving the range of the x (fermi energy) axis. This may need to be
            set manually when including transition level labels, so that they don't cross the axes.
        ylim:
            Tuple (min,max) giving the range for the formation energy axis. This may need to be
            set manually when including transition level labels, so that they don't cross the axes.
        ax_fontsize:
            float  multiplier to change axis label fontsize
        lg_fontsize:
            float  multiplier to change legend label fontsize
        lg_position:
            Tuple (horizontal-position, vertical-position) giving the position
            to place the legend.
            Example: (0.5,-0.75) will likely put it below the x-axis.
    Returns:
        a matplotlib object
    """
    if chempot_limits and "facets" in chempot_limits:
        if not pd_facets:
            pd_facets = chempot_limits[
                "facets"
            ].keys()  # Phase diagram facets to use for chemical
            # potentials, to calculate and plot formation energies
        for facet in pd_facets:
            mu_elts = chempot_limits["facets"][facet]
            elt_refs = chempot_limits["elemental_refs"]
            plot_filename = filename
            if title:
                plot_title = title
                if not filename:
                    plot_filename = plot_title + "_" + facet + ".pdf"
            else:
                plot_title = facet

            return _all_lines_aide_pmg_plot(
                defect_phase_diagram,
                mu_elts=mu_elts,
                elt_refs=elt_refs,
                ax=ax,
                fonts=fonts,
                xlim=xlim,
                ylim=ylim,
                ax_fontsize=ax_fontsize,
                lg_fontsize=lg_fontsize,
                lg_position=lg_position,
                fermi_level=fermi_level,
                title=plot_title,
                saved=saved,
                colormap=colormap,
                frameon=frameon,
                chempot_table=chempot_table,
                auto_labels=auto_labels,
                filename=plot_filename,
            )
    else:  # If you only want to give {Elt: Energy} dict for chempot_limits, or no chempot_limits
        return _all_lines_aide_pmg_plot(
            defect_phase_diagram,
            mu_elts=chempot_limits,
            elt_refs=elt_refs,
            ax=ax,
            fonts=fonts,
            xlim=xlim,
            ylim=ylim,
            ax_fontsize=ax_fontsize,
            lg_fontsize=lg_fontsize,
            lg_position=lg_position,
            fermi_level=fermi_level,
            title=title,
            saved=saved,
            colormap=colormap,
            frameon=frameon,
            chempot_table=chempot_table,
            auto_labels=auto_labels,
            filename=filename,
        )


def _all_lines_aide_pmg_plot(
    defect_phase_diagram,
    mu_elts=None,
    elt_refs=None,
    ax=None,
    fonts=None,
    xlim=None,
    ylim=None,
    ax_fontsize=1.0,
    lg_fontsize=1.0,
    lg_position=None,
    fermi_level=None,
    title=None,
    saved=False,
    colormap="Dark2",
    frameon=False,
    chempot_table=True,
    auto_labels=False,
    filename=None,
):
    """
    Produce defect Formation energy vs Fermi energy plot
    Args:
        mu_elts:
            a dictionary of {Element:value} giving the chemical potential of each element
        xlim:
            Tuple (min,max) giving the range of the x (fermi energy) axis. This may need to be
            set manually when including transition level labels, so that they don't cross the axes.
        ylim:
            Tuple (min,max) giving the range for the formation energy axis. This may need to be
            set manually when including transition level labels, so that they don't cross the axes.
        ax_fontsize:
            float  multiplier to change axis label fontsize
        lg_fontsize:
            float  multiplier to change legend label fontsize
        lg_position:
            Tuple (horizontal-position, vertical-position) giving the position
            to place the legend.
            Example: (0.5,-0.75) will likely put it below the x-axis.
    Returns:
        a matplotlib object
    """

    if xlim is None:
        xlim = (-0.4, defect_phase_diagram.band_gap + 0.4)
    xy = {}
    lower_cap = -100.0
    upper_cap = 100.0
    y_range_vals = []  # for finding max/min values on y-axis based on x-limits

    legends_txt = []
    for chg_ent in defect_phase_diagram.entries:
        defnom = chg_ent.name + f"_{chg_ent.charge}"
        flds = defnom.split("_")
        if flds[0] == "Vac":
            base = r"$\mathrm{V"
            sub_str = "_{" + flds[1] + "}}$"
        elif flds[0] == "Sub":
            flds = defnom.split("_")
            base = r"$\mathrm{" + flds[1]
            sub_str = "_{" + flds[3] + "}}$"
        elif flds[0] == "Int":
            base = r"$\mathrm{" + flds[1]
            sub_str = "_{i}}$"
        else:
            base = defnom
            sub_str = ""
        def_name = (
            base
            + sub_str
            + r"$^{"
            + f"{int(flds[-1]):{'+' if int(flds[-1]) > 0 else ''}}"
            + r"}$"
        )

        # add subscript labels for different configurations of same defect species
        legends_txt = [
            def_name + r"$_{, 1}$" if def_name == orig_def_name else orig_def_name
            for orig_def_name in legends_txt
        ]
        labelled_def_name = def_name + r"$_{, 1}$"
        if labelled_def_name in legends_txt:
            i = 1
            while labelled_def_name in legends_txt:
                i += 1
                labelled_def_name = def_name + r"$_{, " + f"{i}" + r"}$"
            def_name = labelled_def_name
        legends_txt.append(def_name)

        xy[def_name] = [[], []]
        for x_extrem in [lower_cap, upper_cap]:
            xy[def_name][0].append(x_extrem)
            xy[def_name][1].append(
                chg_ent.formation_energy(
                    chemical_potentials=mu_elts, fermi_level=x_extrem
                )
            )
        for x_window in xlim:
            y_range_vals.append(
                chg_ent.formation_energy(
                    chemical_potentials=mu_elts, fermi_level=x_window
                )
            )

    cmap = cm.get_cmap(colormap)
    colors = cmap(np.linspace(0, 1, len(xy)))
    if colormap == "Dark2" and len(xy) >= 8:
        warnings.warn(
            f"""
The chosen colormap is Dark2, which only has 8 colours, yet you have {len(xy)} defect species (so
some defects will have the same line colour). Recommended to change/set colormap to 'tab10' or
'tab20' (10 and 20 colours each)."""
        )
    plt.figure(dpi=600, figsize=(2.6, 1.95))  # Gives a final figure width of c. 3.5
    # inches, the standard single column width for publication (which is what we're about)
    plt.clf()
    width = 9
    ax = pretty_axis(ax=ax, fonts=fonts)
    # plot formation energy lines

    for cnt, def_name in enumerate(xy.keys()):
        ax.plot(
            xy[def_name][0],
            xy[def_name][1],
            color=colors[cnt],
            markeredgecolor=colors[cnt],
            lw=1.2,
            markersize=3.5,
        )

    if not lg_position:
        ax.legend(
            legends_txt,
            fontsize=lg_fontsize * width,
            loc=2,
            bbox_to_anchor=(1, 1),
            frameon=frameon,
            prop=fonts,
        )
    else:
        ax.legend(
            legends_txt,
            fontsize=lg_fontsize * width,
            ncol=3,
            loc="lower center",
            bbox_to_anchor=lg_position,
        )

    if ylim is None:
        window = max(y_range_vals) - min(y_range_vals)
        spacer = 0.1 * window
        ylim = (0, max(y_range_vals) + spacer)
        if auto_labels:  # need to manually set xlim or ylim if labels cross axes!!
            ylim = (0, max(y_range_vals) * 1.17) if spacer / ylim[1] < 0.145 else ylim
            # Increase y_limit to give space for transition level labels

    # Show colourful band edges
    ax.imshow(
        [(0, 1), (0, 1)],
        cmap=plt.cm.Blues,
        extent=(xlim[0], 0, ylim[0], ylim[1]),
        vmin=0,
        vmax=3,
        interpolation="bicubic",
        rasterized=True,
        aspect="auto",
    )

    ax.imshow(
        [(1, 0), (1, 0)],
        cmap=plt.cm.Oranges,
        extent=(defect_phase_diagram.band_gap, xlim[1], ylim[0], ylim[1]),
        vmin=0,
        vmax=3,
        interpolation="bicubic",
        rasterized=True,
        aspect="auto",
    )

    ax.set_ylim(ylim)
    ax.set_xlim(xlim)
    # ax.plot([xlim[0], xlim[1]], [0, 0], "k-")  # black dashed line for E_formation = 0

    if fermi_level is not None:
        plt.axvline(
            x=fermi_level, linestyle="-.", color="k", linewidth=1
        )  # smaller dashed lines for gap edges
    ax.set_xlabel("Fermi Level (eV)", size=ax_fontsize * width)
    ax.set_ylabel("Formation Energy (eV)", size=ax_fontsize * width)
    ax.xaxis.set_major_locator(ticker.MaxNLocator(4))
    ax.xaxis.set_minor_locator(ticker.AutoMinorLocator(2))
    ax.yaxis.set_major_locator(ticker.MaxNLocator(4))
    ax.yaxis.set_minor_locator(ticker.AutoMinorLocator(2))
    ax.xaxis.set_major_formatter(_CustomScalarFormatter(minus_symbol="-"))
    ax.yaxis.set_major_formatter(_CustomScalarFormatter(minus_symbol="-"))
    if chempot_table:
        if elt_refs is not None:
            facets_wrt_elt_refs = {
                elt: energy - elt_refs[elt] for elt, energy in mu_elts.items()
            }
            _plot_chemical_potential_table(
                plt,
                facets_wrt_elt_refs,
                "",
                fontsize=ax_fontsize * width,
                wrt_elt_refs=True,
            )
        elif mu_elts:
            _plot_chemical_potential_table(
                plt,
                mu_elts,
                "",
                fontsize=ax_fontsize * width,
                wrt_elt_refs=False,
            )

    if title and chempot_table:
        ax.set_title(
            latexify(title),
            size=1.2 * ax_fontsize * width,
            pad=28,
            fontdict={"fontweight": "bold"},
        )
    elif title:
        ax.set_title(
            latexify(title), size=ax_fontsize * width, fontdict={"fontweight": "bold"}
        )

    if saved or filename:
        if filename:
            plt.savefig(filename, bbox_inches="tight", dpi=600)
        else:
            plt.savefig(str(title) + "_doped_plot.pdf", bbox_inches="tight", dpi=600)

    return ax



#!/usr/bin/env python

"""
This module is wraparound to sxdefectalign code by Freysoldt
Applies Freysoldt correction with planar averaged potential method

(NOTE from PyCDT v2.0 transition) This code is not used explicitly
by PyCDT script, but there are no plans to merge this to pymatgen
in the future (since equivalent functionality exists there).
Keeping code here in case someone wants to make use of future
sxdefectalign functionalities with a python wrapper.
"""

__status__ = "Development"

import os
import subprocess

import numpy as np
from monty.tempfile import ScratchDir
from pymatgen.io.vasp.outputs import Locpot

from doped.pycdt.utils.parse_calculations import get_locpot


class SxdefectalignWrapper(object):
    """
        NOTE from developers:
            This code is a python wrapper to sxdefectalign.
            We have our own version written in python within
                pycdt.corrections.freysoldt_correction
            This code has no unit test and will not be maintained
            going forward (as of 12/15/2017).
            However, we are keeping function here to allow for
            current users to make use of it...
    This class applies the Freysoldt correction code (sxdefectalign)
    to remove electrostatic defect
    interaction contribution to energy and apply potential alignment.
    """

    def __init__(
        self,
        locpot_bulk_path,
        locpot_defect_path,
        charge,
        epsilon,
        site_frac_coords,
        encut,
        lengths=None,
        name="",
    ):
        """
        Args:
            locpot_bulk:
                Location of LOCPOT of bulk
            locpot_defect:
                Location of LOCPOT of defect
            charge:
                Charge relative to neutral defect (not relative to bulk)
            epsilon:
                Dielectric constant obtained from relaxation run
            site_frac_coords:
                Fractional coordinates of defect site as list
            encut:
                Planewave basis energy cutoff used in VASP run (in eV)
            name:
                Name of the defect to write files
            lengths:
                Length of lattice vectors.
        """
        self._locpot_bulk = locpot_bulk_path
        self._locpot_defect = locpot_defect_path
        self._charge = charge
        self._epsilon = epsilon
        self._frac_coords = site_frac_coords
        self._encut = encut
        if not lengths:
            struct = get_locpot(locpot_bulk_path)
            self._lengths = struct.structure.lattice.abc
            print(
                "had to import lengths, if want to speed up set lengths="
                + str(self._lengths)
            )
        else:
            self._lengths = lengths
        self._name = name

    def prepare_files(self):
        if self._charge == 0:
            print("defect has charge 0, so freysoldt correction is 0")
            return
        # self._locpot_bulk.write_file("LOCPOT_vref",True)   #from when these were locpot objects
        # self._locpot_defect.write_file("LOCPOT_vdef",True) #from when these were locpot objects
        path, name = os.path.split(self._locpot_bulk)
        mod_blk_locpot = os.path.join(path, name + "_vref")
        self.mod_bulk_locpot = mod_blk_locpot
        if not os.path.exists(mod_blk_locpot):
            print("prep pure Locpot")
            with open(self._locpot_bulk) as input:
                with open(mod_blk_locpot, "w") as output:
                    cnt = 0
                    for line in input:
                        cnt += 1
                        if cnt != 6:
                            output.write(line)
            # cmd="perl -n -e 'print if $. != 6' "+str(self._locpot_bulk)+" > "+mod_blk_locpot
            # print cmd
            # os.system(cmd)
        path, name = os.path.split(self._locpot_defect)
        mod_defect_locpot = os.path.join(path, name + "_vdef")
        self.mod_defect_locpot = mod_defect_locpot
        if not os.path.exists(mod_defect_locpot):
            print("prep defect Locpot")
            with open(self._locpot_defect) as input:
                with open(mod_defect_locpot, "w") as output:
                    cnt = 0
                    for line in input:
                        cnt += 1
                        if cnt != 6:
                            output.write(line)
            # md="perl -n -e 'print if $. != 6' "+str(self._locpot_defect)+" > "+mod_def_locpot
            # rint cmd
            # s.system(cmd)
        print("locpots prepared for sxdefectalign")

    def plot_hartree_pot(self):
        # plot planar averages of bulk and defect (good for seeing global changes)
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(3, 1, 1)
        ax.set_title("Locpot planar averaged potentials")
        bulkloc = get_locpot(self._locpot_bulk)
        defloc = get_locpot(self._locpot_bulk)
        get_agrid = bulkloc.get_axis_grid
        get_baavg = bulkloc.get_average_along_axis
        get_daavg = defloc.get_average_along_axis
        for axis in [0, 1, 2]:
            ax = fig.add_subplot(3, 1, axis + 1)
            latt_len = self._lengths[axis]
            ax.plot(get_agrid(axis), get_baavg(axis), "r", label="Bulk potential")
            ax.plot(get_agrid(axis), get_daavg(axis), "b", label="Defect potential")
            ax.plot(
                [self._frac_coords[axis] * latt_len],
                [0],
                "or",
                markersize=4.0,
                label="Defect site",
            )
            ax.set_ylabel("axis " + str(axis + 1))
            if not axis:
                ax.legend()
        ax.set_xlabel("distance (Angstrom)")
        # plt.savefig('locpotavgplot.png')
        plt.show()

    def plot_hartree_pot_diff(self):
        # only plot the difference in planar averaged potentials
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(3, 1, 1)
        ax.set_title("Locpot planar averaged potential difference")
        bulkloc = get_locpot(self._locpot_bulk)
        defloc = get_locpot(self._locpot_defect)
        for axis in [0, 1, 2]:
            ax = fig.add_subplot(3, 1, axis + 1)
            defect_axis = defloc.get_axis_grid(axis)
            defect_pot = defloc.get_average_along_axis(axis)
            pure_pot = bulkloc.get_average_along_axis(axis)
            latt_len = self._lengths[axis]
            ax.plot(
                defect_axis, defect_pot - pure_pot, "b", label="Defect-Bulk difference"
            )
            ax.plot(
                [self._frac_coords[axis] * latt_len],
                [0],
                "or",
                markersize=4.0,
                label="Defect site",
            )
            ax.set_ylabel("axis " + str(axis + 1))
            if not axis:
                ax.legend()
        ax.set_xlabel("distance (Angstrom)")
        # plt.savefig('locpotdiffplot.png')
        plt.show()

    def plot_all_hartree_pot(self):
        # plot planar averaged locpots, along with the difference for each axis
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(3, 1, 1)
        ax.set_title("Locpot planar averaged potentials and difference")
        bulkloc = get_locpot(self._locpot_bulk)
        defloc = get_locpot(self._locpot_defect)
        for axis in [0, 1, 2]:
            ax = fig.add_subplot(3, 1, axis + 1)
            defect_axis = defloc.get_axis_grid(axis)
            defect_pot = defloc.get_average_along_axis(axis)
            pure_pot = bulkloc.get_average_along_axis(axis)
            ax.plot(defect_axis, pure_pot, "r", label="Bulk potential")
            ax.plot(defect_axis, defect_pot, "b", label="Defect potential")
            ax.plot(
                defect_axis, defect_pot - pure_pot, "k", label="Defect-Bulk difference"
            )
            ax.plot(
                [self._frac_coords[axis] * self._lengths[axis]],
                [0],
                "or",
                markersize=4.0,
                label="Defect site",
            )
            ax.set_ylabel("axis " + str(axis + 1))
            if not axis:
                ax.legend()
        ax.set_xlabel("distance (Angstrom)")
        # plt.savefig('locpotavgdiffplot.png')
        plt.show()

    def plot_pot_diff(self, align=[0.0, 0.0, 0.0], print_pot_flag="written"):
        """
        Runs sxdefectalign and obtains three alignment constants for
        each axis. To obtain Freysoldt correction call this function twice
        (once to get alignment, and second time to get final correction
        and plots). In the first call, use print_pot_flag=none.
        Then in second call get corrections and leave vline-eV.dat
        files for each axis and produce potential alignment correction
        or if  print flag= written then get different parts of vline-eV.dat
        written to files for plotting (good for nersc or where you can't
        plot easy) or if print flag=plotfull then show the full matplotlib
        plot for all three axes at end of calculation.
        encut is eV cutoff from VASP

        Would like final workflow to include the flag that I have put here
        for when planar average varies by more than 0.2 eV around far region
        """
        if not self._charge:  # don't need charge correction if charge is zero
            return [[0, 0, 0], [0, 0, 0]]
        # correction from output (should include alignment once alignment
        # has been done)
        result = []
        platy = []  # alignment terms for each axis
        # if want to plot right here, then build dictionary for storing
        # planar average values of each axis
        if print_pot_flag == "plotfull":
            plotvals = {"0": {}, "1": {}, "2": {}}
        for axis in [0, 1, 2]:
            print("do axis " + str(axis + 1))
            # print self._frac_coords[1:]
            # relpos = (str(self._frac_coords)[1:])[:-1]
            # relpos = relpos.replace(" ",",")
            relpos = ",".join(str(i) for i in self._frac_coords)
            command = [
                "sxdefectalign",
                "--vasp",
                "-a" + str(axis + 1),
                "--relative",
                "--pos",
                relpos,
                "--charge",
                str(-self._charge),
                "--ecut",
                str(self._encut / 13.6057),  # eV to Ry for sxdefect
                "--eps",
                str(self._epsilon),
                "-C",
                str(-float(align[axis])),
                "--vref",
                self.mod_bulk_locpot,
                "--vdef",
                self.mod_defect_locpot,
            ]
            print(command)

            # standard way of running NERSC commands.
            # in case NERSC (Hopper) has issues with python subprocess can use hack
            # p = subprocess.Popen(command, stdout=subprocess.PIPE,
            #        stdin=subprocess.PIPE, stderr=subprocess.PIPE)
            # output, err = p.communicate()
            # out = out.decode('utf-8')
            # err = err.decode('utf-8')
            # print 'output from sxdefectalign = ', str(out)
            # result.append(float(out[0].split("\n")[12].split()[3]))
            # print "chg correction is "+str(result[-1])

            # this is hack wrap-around for when subprocess doesn't work
            # (which is always an issue on hopper now...)
            cmd = " ".join(command)
            cmd += " > tmpoutput"
            os.system(cmd)

            with open("tmpoutput") as f:
                output = f.readlines()
            # print 'output from sxdefectalign = '+str(output)
            val = output[-1].split()[3].strip()
            # result.append(float(output[-1].split()[3]))
            result.append(float(val))
            print("chg correction is " + str(result[-1]))
            os.remove("tmpoutput")

            x_lr, y_lr = [], []
            x, y = [], []
            x_diff, y_diff = [], []
            with open("vline-eV.dat", "r") as f_sr:  # read in potential
                for r in f_sr:
                    tmp = r.split("\t")
                    if len(tmp) < 3 and not r.startswith("&"):
                        x_lr.append(float(tmp[0]) / 1.889725989)  # to Angstrom
                        y_lr.append(float(tmp[1]))
                    if len(tmp) > 2:
                        x.append(float(tmp[0]) / 1.889725989)  # to Angstrom
                        x_diff.append(float(tmp[0]) / 1.889725989)  # to Angstrom
                        y.append(float(tmp[2].rstrip("\n")))
                        y_diff.append(float(tmp[1]))

            if print_pot_flag != "none":
                if os.path.exists("vline-eV.dat"):
                    os.rename("vline-eV.dat", "axis" + str(axis) + "vline-eV.dat")
            else:
                if os.path.exists("vline-eV.dat"):
                    os.remove("vline-eV.dat")

            # Extract potential alignment term averaging window of +/- 1 Ang
            # around point halfway between neighboring defects
            latt_len = self._lengths[axis]
            if self._frac_coords[axis] >= 0.5:
                platx = (self._frac_coords[axis] - 0.5) * latt_len
            else:
                platx = (self._frac_coords[axis] + 0.5) * latt_len
            print("half way between defects is: ", platx)

            sample_radius = 0.5  # 1 / 2
            xmin = (
                latt_len - (sample_radius - platx)
                if platx < sample_radius
                else platx - sample_radius
            )
            xmax = (
                sample_radius - (latt_len - platx)
                if platx > latt_len - sample_radius
                else sample_radius + platx
            )
            print("means sampling region is (", xmin, ",", xmax, ")")

            tmpalign = []
            if xmax < xmin:
                print("wrap around detected, special alignment needed")
                for i in range(len(x)):
                    if x[i] < xmax or x[i] > xmin:
                        tmpalign.append(y[i])
                    else:
                        continue
            else:
                for i in range(len(x)):
                    if x[i] > xmin and x[i] < xmax:
                        tmpalign.append(y[i])
                    else:
                        continue

            print("alignment is ", -np.mean(tmpalign))
            platy.append(-np.mean(tmpalign))
            flag = 0
            for i in tmpalign:  # check to see if alignment region varies too much
                if np.abs(i - platy[-1]) > 0.2:
                    flag = 1
                else:
                    continue
            if flag != 0:
                print(
                    "Warning: potential aligned region varied by more "
                    + "than 0.2eV (in range of halfway between defects "
                    + "+/-1 \Angstrom). Might have issues with Freidel "
                    + "oscillations or atomic relaxation"
                )

            if print_pot_flag == "written":

                def write_xy(x, y, fname):
                    """
                    Write the x, y vectors to file
                    """
                    with open(os.path.join("..", fname), "w") as f:
                        for pair in zip(x, y):
                            print >> f, " ".join(str(i) for i in pair)

                name = self._name
                charge = str(self._charge)
                fname = "_".join([name, charge, "xylong", str(axis)]) + ".dat"
                write_xy(x_lr, y_lr, fname)
                fname = "_".join([name, charge, "xy", str(axis)]) + ".dat"
                # fname = name + charge + 'xy' + str(axis) + '.dat'
                write_xy(x, y, fname)
                fname = "_".join([name, charge, "xy", str(axis), "diff.dat"])
                # fname = name + charge + 'xy' + str(axis) + 'diff.dat'
                write_xy(x_diff, y_diff, fname)

            elif (
                print_pot_flag == "plotfull"
            ):  # store data for plotting at end of all calcs
                plotvals[str(axis)]["xylong"] = [x_lr, y_lr]
                plotvals[str(axis)]["xy"] = [x, y]
                plotvals[str(axis)]["xydiff"] = [x_diff, y_diff]

        if print_pot_flag == "plotfull":  # plot all three planar averaged potentials
            import matplotlib.pyplot as plt
            import pylab

            fig = plt.figure(figsize=(15.0, 12.0))
            print("plot full plot")
            for axis in [0, 1, 2]:
                print("plot axis ", axis + 1)
                ax = fig.add_subplot(3, 1, axis + 1)
                ax.set_ylabel("axis " + str(axis + 1))
                # pylab.hold(True)
                vals_plot = plotvals[str(axis)]
                ax.plot(vals_plot["xy"][0], vals_plot["xy"][1])
                ax.plot(vals_plot["xydiff"][0], vals_plot["xydiff"][1], "r")
                ax.plot(vals_plot["xylong"][0], vals_plot["xylong"][1], "g")

                if not axis:
                    ax.set_title("Electrostatic planar averaged potential")
                    ax.legend(["V_defect-V_ref-V_lr", "V_defect-V_ref", "V_lr"])

                latt_len = self._lengths[axis]
                fcoords = self._frac_coords[axis]
                ax.plot([fcoords * latt_len], [0], "or", markersize=4.0)
                if fcoords >= 0.5:
                    ax.plot([(fcoords - 0.5) * latt_len], [0], "og", markersize=4.0)
                else:
                    ax.plot([(fcoords + 0.5) * latt_len], [0], "og", markersize=4.0)
                plt.axhline(y=0.0, linewidth=0.8, color="black")
            plt.savefig(os.path.join("..", "locpotgraph.png"))

        return [result, platy]

    def run_correction(self, print_pot_flag="written", partflag="All"):
        """
        Runs all neccessary parts to get freysoldt corrections out with
        planar averaged potentials
        Change plot_pot_flag if you want plotted planar averages
        set transflag to True if you want to write flags
        """
        outputvals = []  # for splitting up parts of correction
        # with ScratchDir('.'):
        self.prepare_files()
        s = self.plot_pot_diff(print_pot_flag="none")
        outputvals.append(np.mean(s[0]))  # ES correction
        outputvals.append(-self._charge * np.mean(s[1]))
        print("--")
        print("potential alignments determined to be: " + str(s[1]))
        print("get final correction terms")
        print("--")
        # To get locpot plots use print_pot_flag = 'written' or 'plotfull'
        vals = self.plot_pot_diff(align=s[1], print_pot_flag=print_pot_flag)
        print("vals is " + str(vals))
        for i in range(3):
            if np.abs(vals[1][i]) > 0.0001:
                print(
                    "PROBLEM! planar averaging didnt work. Issue " + "with axis",
                    str(i + 1),
                )
        print("--")
        print("Final Freysoldt sxdefectalign correction values are: ", str(vals[0]))
        outputvals.append(np.mean(vals[0]))

        if partflag == "All":
            return outputvals[2]
        elif partflag == "AllSplit":
            return outputvals
        elif partflag == "pc":
            return outputvals[0]
        elif partflag == "potalign":
            return outputvals[1]

        return None


if __name__ == "__main__":
    pass



# coding: utf-8
"""
A class for performing analysis of chemical potentials with the grand
canonical linear programming approach
"""
from __future__ import division

import copy
import logging
import os

from pymatgen.analysis.phase_diagram import PDEntry, PhaseDiagram
from pymatgen.core.structure import Element, Structure
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.ext.matproj import MPRester


def get_mp_chempots_from_dpd(dpd):
    """
    Grab Materials Project chemical potentials from a pymatgen DefectPhaseDiagram object
    """
    print("Retrieving chemical potentials from MP database using dpd object...")
    bulk_energy = 0.0
    if ("bulk_energy" in dpd.entries[0].parameters.keys()) and (
        "bulk_sc_structure" in dpd.entries[0].parameters.keys()
    ):
        try:
            bulk_struct = dpd.entries[0].parameters["bulk_sc_structure"].copy()
            if not isinstance(bulk_struct, Structure):
                bulk_struct = Structure.from_dict(bulk_struct)
            bulk_energy = dpd.entries[0].parameters["bulk_energy"]
        except:
            print(
                "Failure in grabbing bulk energy and structure for chemical potential parsing."
            )
    if not bulk_energy:
        print(
            "Grabbing chemical potentials without analyzing stability of bulk structure. "
            "Ignore any flags raised about stability of structure"
        )
        bulk_energy = 0.0
        bulk_struct = dpd.entries[0].defect.bulk_structure.copy()

    bulk_ce = ComputedStructureEntry(bulk_struct, bulk_energy)
    bulk_elt_set = list(bulk_struct.symbol_set)

    sub_species = []
    for entry in dpd.entries:
        def_site = entry.defect.site.specie.symbol
        if def_site not in bulk_elt_set:
            sub_species.append(def_site)

    sub_species = set(sub_species)
    print("Bulk symbols = {}, Sub symbols = {}".format(bulk_elt_set, sub_species))
    mp_cpa = MPChemPotAnalyzer(bulk_ce=bulk_ce, sub_species=sub_species)

    return mp_cpa.analyze_GGA_chempots()


class ChemPotAnalyzer:
    """
    Post processing for atomic chemical potentials used in defect
    calculations.
    """

    def __init__(self, **kwargs):
        """
        Args:
            bulk_ce: Pymatgen ComputedStructureEntry object for
                bulk entry / supercell
        """
        self.bulk_ce = kwargs.get("bulk_ce", None)

    def get_chempots_from_pd(self, pd):
        logger = logging.getLogger(__name__)

        if not self.bulk_ce:
            msg = (
                "No bulk entry supplied. "
                "Cannot compute atomic chempots without knowing the bulk entry of interest."
            )
            logger.warning(msg)
            raise ValueError(msg)

        bulk_composition = self.bulk_ce.composition
        redcomp = bulk_composition.reduced_composition
        # append bulk_ce to phase diagram, if not present
        entries = pd.all_entries
        if not any(
            [
                (
                    ent.composition == self.bulk_ce.composition
                    and ent.energy == self.bulk_ce.energy
                )
                for ent in entries
            ]
        ):
            entries.append(
                PDEntry(
                    self.bulk_ce.composition,
                    self.bulk_ce.energy,
                    attribute="Bulk Material",
                )
            )
            pd = PhaseDiagram(entries)

        chem_lims = pd.get_all_chempots(redcomp)

        return chem_lims

    def diff_bulk_sub_phases(self, face_list, sub_el=None):
        # method for pulling out phases within a facet of a phase diagram
        # which may include a substitutional element...
        # face_list is an array of phases in a facet
        # sub_el is the element to look out for within the face_list array
        blk = []
        sub_spcs = []
        for face in face_list:
            if sub_el:
                if sub_el in face:
                    sub_spcs.append(face)
                else:
                    blk.append(face)
            else:
                blk.append(face)
        blk.sort()
        sub_spcs.sort()
        blknom = "-".join(blk)
        subnom = "-".join(sub_spcs)
        return blk, blknom, subnom


class MPChemPotAnalyzer(ChemPotAnalyzer):
    """
    Post processing for atomic chemical potentials by querying MP database

    Makes use of Materials Project pre-computed data to generate
    needed information for chem pots in different growth conditions.

    WARNING: If you plan to use this method, then you better be sure you are
    using the same settings as MP (same INCAR, POTCARs etc.)
    """

    def __init__(self, **kwargs):
        """
        Args:
            bulk_ce: Pymatgen ComputedStructureEntry object for
                bulk entry / supercell
            sub_species (set): set of elemental species that are extrinsic to structure.
                Default is no subs included
            entries (dict): a dict of pymatgen ComputedEntry objects to build relevant phase diagram
                The dict contains two keys: 'bulk_derived', and 'subs_set', each contains a list
                of ComputedEntry objects
                'bulk_derived' list only has compositions containing elements from the bulk (
                un-defective) composition
                'subs_set' list has compositions which contain at least one element that is not
                in the bulk composition
            mpid (str): Materials Project ID of bulk structure (not required, can use bulk_ce
            instead);
                format "mp-X", where X is an integer;
            mapi_key (str): Materials API key to access database
                (if not in ~/.pmgrc.yaml already)
        """
        super(self.__class__, self).__init__(**kwargs)
        self.sub_species = kwargs.get("sub_species", set())
        self.entries = kwargs.get("entries", {})
        self.mpid = kwargs.get("mpid", None)
        self.mapi_key = kwargs.get("mapi_key", None)

    def analyze_GGA_chempots(self, full_sub_approach=False):
        """
        For calculating GGA-PBE atomic chemical potentials by using
            Materials Project pre-computed data

        Args:
            full_sub_approach: generate chemical potentials by looking at
                full phase diagram (setting to True is NOT recommended
                if subs_species set has more than one element in it...)

        This code retrieves atomic chempots from Materials
        Project (MP) entries by making use of the pymatgen
        phase diagram (PD) object and computed entries from the MP
        database. There are debug notes that are made based on the stability of
        the structure of interest with respect to the phase diagram generated from MP

        NOTE on 'full_sub_approach':
            The default approach for substitutional elements (full_sub_approach = False)
            is to only consider facets which have a maximum of 1 composition with
            the extrinsic species present
            (see PyCDT paper for chemical potential methodology DOI: 10.1016/j.cpc.2018.01.004).

            This default approach speeds up analysis when analyzing several substitutional
            species at the same time.

            If you prefer to consider the full phase diagram (not recommended
            when you have more than 2 substitutional defects), then set
            full_sub_approach to True.
        """
        logger = logging.getLogger(__name__)

        # gather entries
        self.get_mp_entries(full_sub_approach=full_sub_approach)

        # figure out how system should be treated for chemical potentials
        # based on phase diagram
        entry_list = self.entries["bulk_derived"]
        pd = PhaseDiagram(entry_list)

        decomp_en = round(
            pd.get_decomp_and_e_above_hull(self.bulk_ce, allow_negative=True)[1], 4
        )

        stable_composition_exists = False
        for i in pd.stable_entries:
            if i.composition.reduced_composition == self.redcomp:
                stable_composition_exists = True

        if (decomp_en <= 0) and stable_composition_exists:
            logger.debug(
                "Bulk Computed Entry found to be stable with respect "
                "to MP Phase Diagram (e_above_hull = {} eV/atom).".format(decomp_en)
            )
        elif (decomp_en <= 0) and not stable_composition_exists:
            logger.info(
                "Bulk Computed Entry found to be stable with respect "
                "to MP Phase Diagram (e_above_hull = {} eV/atom).\n"
                "However, no stable entry with this composition exists "
                "in the MP database!\nPlease consider submitting the "
                "POSCAR to the MP xtaltoolkit, so future users will "
                "know about this structure:"
                " https://materialsproject.org/#apps/xtaltoolkit\n"
                "Manually inserting structure into phase diagram and "
                "proceeding as normal.".format(decomp_en)
            )
            entry_list.append(self.bulk_ce)
        elif stable_composition_exists:
            logger.warning(
                "Bulk Computed Entry not stable with respect to MP "
                "Phase Diagram (e_above_hull = {} eV/atom), but found "
                "stable MP composition to exist.\nProducing chemical "
                "potentials with respect to stable phase.".format(decomp_en)
            )
        else:
            logger.warning(
                "Bulk Computed Entry not stable with respect to MP "
                "Phase Diagram (e_above_hull = {} eV/atom) and no "
                "stable structure with this composition exists in the "
                "MP database.\nProceeding with atomic chemical "
                "potentials according to composition position within "
                "phase diagram.".format(decomp_en)
            )

        pd = PhaseDiagram(entry_list)
        chem_lims = self.get_chempots_from_pd(pd)
        logger.debug("Bulk Chemical potential facets: {}".format(chem_lims.keys()))

        if not full_sub_approach:
            # NOTE if full_sub_approach was True, then all the sub_entries
            # would be ported into the bulk_derived list
            finchem_lims = {}  # this will be final chem_lims dictionary
            for key in chem_lims.keys():
                face_list = key.split("-")
                blk, blknom, subnom = self.diff_bulk_sub_phases(face_list)
                finchem_lims[blknom] = {}
                finchem_lims[blknom] = chem_lims[key]

            # Now add single elements to extend the phase diagram,
            # adding new additions to chemical potentials ONLY for the cases
            # where the phases in equilibria are those from the bulk phase
            # diagram. This is essentially the assumption that the majority of
            # the elements in the total composition will be from the native
            # species present rather than the sub species (a good approximation)
            for sub_el in self.sub_species:
                sub_specie_entries = entry_list[:]
                for entry in self.entries["subs_set"][sub_el]:
                    sub_specie_entries.append(entry)

                pd = PhaseDiagram(sub_specie_entries)
                chem_lims = self.get_chempots_from_pd(pd)

                for key in chem_lims.keys():
                    face_list = key.split("-")
                    blk, blknom, subnom = self.diff_bulk_sub_phases(
                        face_list, sub_el=sub_el
                    )
                    # if number of facets from bulk phase diagram is
                    # equal to bulk species then full_sub_approach says this
                    # can be grouped with rest of structures
                    if len(blk) == len(self.bulk_species_symbol):
                        if blknom not in finchem_lims.keys():
                            finchem_lims[blknom] = chem_lims[key]
                        else:
                            finchem_lims[blknom][Element(sub_el)] = chem_lims[key][
                                Element(sub_el)
                            ]
                        if "name-append" not in finchem_lims[blknom].keys():
                            finchem_lims[blknom]["name-append"] = subnom
                        else:
                            finchem_lims[blknom]["name-append"] += "-" + subnom
                    else:
                        # if chem pots determined by two (or more) sub-specie
                        # containing phases, skip this facet!
                        continue
            # run a check to make sure all facets dominantly defined by bulk species
            overdependent_chempot = False
            facets_to_delete = []
            for facet_name, cps in finchem_lims.items():
                cp_key_num = (
                    (len(cps.keys()) - 1) if "name-append" in cps else len(cps.keys())
                )
                if cp_key_num != (
                    len(self.bulk_species_symbol) + len(self.sub_species)
                ):
                    facets_to_delete.append(facet_name)
                    logger.info(
                        "Not using facet {} because insufficient number of bulk facets for "
                        "bulk set {} with sub_species set {}. (only dependent on {})."
                        "".format(
                            facet_name,
                            self.bulk_species_symbol,
                            self.sub_species,
                            cps.get("name-append"),
                        )
                    )
            if len(facets_to_delete) == len(finchem_lims):
                overdependent_chempot = True
                logger.warning(
                    "Determined chemical potentials to be over dependent"
                    " on a substitutional specie. Needing to revert to full_sub_approach. If "
                    "multiple sub species exist this could take a while/break the code..."
                )
            else:
                finchem_lims = {
                    k: v for k, v in finchem_lims.items() if k not in facets_to_delete
                }

            if not overdependent_chempot:
                chem_lims = {}
                for orig_facet, fc_cp_dict in finchem_lims.items():
                    if "name-append" not in fc_cp_dict:
                        facet_nom = orig_facet
                    else:
                        full_facet_list = orig_facet.split("-")
                        full_facet_list.extend(fc_cp_dict["name-append"].split("-"))
                        full_facet_list.sort()
                        facet_nom = "-".join(full_facet_list)
                    chem_lims[facet_nom] = {
                        k: v for k, v in fc_cp_dict.items() if k != "name-append"
                    }
            else:
                # This is for when overdetermined chempots occur, forcing the full_sub_approach
                # to happen
                for sub, subentries in self.entries["subs_set"].items():
                    for subentry in subentries:
                        entry_list.append(subentry)
                pd = PhaseDiagram(entry_list)
                chem_lims = self.get_chempots_from_pd(pd)

        return chem_lims

    def get_chempots_from_composition(self, bulk_composition, full_sub_approach=False):
        """
        A simple method for getting GGA-PBE chemical potentials JUST
        from the composition information (Note: this only works if the
        composition already exists in the MP database)

        Args:
            bulk_composition : Composition of bulk as a pymatgen Composition
                object. This and mapi_key are only actual required input for
                generating set of chemical potentials from Materials Project
                database
            full_sub_approach : Include sub_species in query, sub entries can
                be found in self.entries['subs_set']
        """
        logger = logging.getLogger(__name__)

        redcomp = bulk_composition.reduced_composition
        self.bulk_species_symbol = [s.symbol for s in redcomp.elements]

        if not self.entries:
            if (
                full_sub_approach
            ):  # this can be time consuming if several sub species exist
                species_symbols = self.bulk_species_symbol[:]
                for sub_el in self.sub_species:
                    species_symbols.append(sub_el)

                with MPRester(api_key=self.mapi_key) as mp:
                    self.entries["bulk_derived"] = mp.get_entries_in_chemsys(
                        species_symbols
                    )

                self.entries["subs_set"] = {sub_el: [] for sub_el in self.sub_species}
                for entry in self.entries["bulk_derived"]:
                    for sub_el in self.sub_species:
                        if sub_el in entry.composition:
                            self.entries["subs_set"][sub_el].append(entry)

            else:
                with MPRester(api_key=self.mapi_key) as mp:
                    self.entries["bulk_derived"] = mp.get_entries_in_chemsys(
                        self.bulk_species_symbol
                    )

        pd = PhaseDiagram(self.entries["bulk_derived"])
        chem_lims = pd.get_all_chempots(redcomp)

        return chem_lims

    def get_mp_entries(self, full_sub_approach=False):
        """
        This queries MP database for computed entries according to
        input bulk and sub elements of interest

        Args:
            mpid (str): Structure id of the system in the MP databse.
            mapi_key (str): Materials API key to access database
                (if not in ~/.pmgrc.yaml already)
        """
        logger = logging.getLogger(__name__)

        if self.bulk_ce:
            self.bulk_species_symbol = [
                s.symbol for s in self.bulk_ce.composition.elements
            ]
            self.redcomp = self.bulk_ce.composition.reduced_composition
            bce_override = True
        elif self.mpid:
            with MPRester(api_key=self.mapi_key) as mp:
                self.bulk_ce = mp.get_entry_by_material_id(self.mpid)
            self.bulk_species_symbol = [
                s.symbol for s in self.bulk_ce.composition.elements
            ]
            self.redcomp = self.bulk_ce.composition.reduced_composition
            bce_override = False
        else:
            msg = (
                "No bulk entry OR mpid supplied. "
                "Cannot compute atomic chempots without know the bulk entry of interest."
            )
            logger.warning(msg)
            raise ValueError(msg)

        if full_sub_approach:  # this can be time consuming if several sub species exist
            species_symbols = self.bulk_species_symbol[:]
            for sub_el in self.sub_species:
                species_symbols.append(sub_el)

            with MPRester(api_key=self.mapi_key) as mp:
                self.entries["bulk_derived"] = mp.get_entries_in_chemsys(
                    species_symbols
                )

            self.entries["subs_set"] = {sub_el: [] for sub_el in self.sub_species}
            for entry in self.entries["bulk_derived"]:
                for sub_el in self.sub_species:
                    if sub_el in entry.composition:
                        self.entries["subs_set"][sub_el].append(entry)

        else:  # this is recommended approach for running sub species seperately (assumes subs
            # are in dilute concentrations)
            with MPRester(api_key=self.mapi_key) as mp:
                self.entries["bulk_derived"] = mp.get_entries_in_chemsys(
                    self.bulk_species_symbol
                )
                if self.mpid and bce_override:  # overriding bulk_ce if mp-id is given.
                    self.bulk_ce = mp.get_entry_by_material_id(self.mpid)
            if not self.entries:
                msg = (
                    "Could not fetch bulk entries for atomic chempots!"
                    "MPRester query error."
                )
                logger.warning(msg)
                raise ValueError(msg)

            # now compile substitution entries
            self.entries["subs_set"] = dict()
            bulk_entry_set = [entry.entry_id for entry in self.entries["bulk_derived"]]
            for sub_el in self.sub_species:
                els = self.bulk_species_symbol + [sub_el]
                with MPRester(api_key=self.mapi_key) as mp:
                    sub_entry_set = mp.get_entries_in_chemsys(els)
                if not sub_entry_set:
                    msg = (
                        "Could not fetch sub entries for {} atomic chempots! "
                        "Encountered MPRester query error".format(sub_el)
                    )
                    logger.warning(msg)
                    raise ValueError(msg)

                fin_sub_entry_set = []
                for entry in sub_entry_set:
                    if entry.entry_id not in bulk_entry_set:
                        fin_sub_entry_set.append(entry)
                # All entries apart from the bulk entry set
                self.entries["subs_set"][sub_el] = fin_sub_entry_set


class UserChemPotAnalyzer(ChemPotAnalyzer):
    """
    Post processing for atomic chemical potentials based on user computed
    phase diagram entries (possibly supplemented with MP database entries)
    """

    def __init__(self, **kwargs):
        """
        Args:
            bulk_ce: Pymatgen ComputedStructureEntry object for bulk entry
                or supercell
            path_base (str): the base path where the 'PhaseDiagram' folder
                exists defaults to the local folder
            sub_species (set): set of elemental species that are extrinsic
                to structure. Default is no subs included
            entries (dict): pymatgen ComputedEntry objects to build phase
                diagram The dict contains two keys: 'bulk_derived', and
                'subs_set', each contains a list of computed entries
                bulk_derived entries only have a composition containing
                elements from the set of elements in the bulk phase
                subs_set contains elements that are extrinsic to the
                structure of interest
            mapi_key (str): Materials API key to access database
                (if not in ~/.pmgrc.yaml already)
        """
        super(self.__class__, self).__init__(**kwargs)
        self.path_base = kwargs.get("path_base", ".")
        self.sub_species = kwargs.get("sub_species", set())
        self.entries = kwargs.get("entries", {})
        self.mapi_key = kwargs.get("mapi_key", None)

    def read_phase_diagram_and_chempots(
        self, full_sub_approach=False, include_mp_entries=True
    ):
        """
        Once phase diagram has been set up and run by user (in a folder
        called "PhaseDiagram"), this method parses and prints the chemical
        potentials based on the computed entries. The methodology is
        basically identical to that in the analyze_GGA_chempots method.

        Will supplement unfinished entries with MP database entries
        unless no_mp_entries is set to False

        Args:
            full_sub_approach: same attribute as described at length in
                the analyze_GGA_chempots method. Basically, the user can
                set this to True if they want to mix extrinsic species
                in the phase diagram

            include_mp_entries: if set to True, extra entries from
                Materials Project will be added to phase diagram
                according to phases that are stable in the Materials
                Project database

        """
        pdfile = os.path.join(self.path_base, "PhaseDiagram")
        if not os.path.exists(pdfile):
            print("Phase diagram file does not exist at ", pdfile)
            return

        # this is where we read computed entries into a list for parsing...
        # NOTE TO USER: If not running with VASP need to use another
        # pymatgen functionality for importing computed entries below...
        personal_entry_list = []
        for structfile in os.listdir(pdfile):
            if os.path.exists(
                os.path.join(pdfile, structfile, "vasprun.xml")
            ) or os.path.exists(os.path.join(pdfile, structfile, "vasprun.xml.gz")):
                try:
                    print("loading ", structfile)
                    from doped.pycdt.utils.parse_calculations import get_vasprun

                    vr, vr_path = get_vasprun(
                        os.path.join(pdfile, structfile, "vasprun.xml")
                    )
                    vr_entry = vr.get_computed_entry()
                    pdentry = PDEntry(
                        vr_entry.composition, vr_entry.energy, attribute=structfile
                    )
                    personal_entry_list.append(pdentry)
                except:
                    print("Could not load ", structfile)

            else:
                print("No vasprun.xml(.gz) found in ", structfile)

        # add bulk computed entry to phase diagram, and see if it is stable
        if not self.bulk_ce:
            vr_path = os.path.join(self.path_base, "bulk", "vasprun.xml")
            if os.path.exists(vr_path):
                print("loading bulk computed entry")
                from doped.pycdt.utils.parse_calculations import get_vasprun

                bulkvr, bulkvr_path = get_vasprun(vr_path)
                bulkvr_entry = bulkvr.get_computed_entry()
                self.bulk_ce = PDEntry(
                    bulkvr_entry.composition, bulkvr_entry.energy, attribute=bulkvr_path
                )
            else:
                print(
                    "No bulk entry given locally. Phase diagram "
                    + "calculations cannot be set up without this"
                )
                return

        self.bulk_composition = self.bulk_ce.composition
        self.redcomp = self.bulk_composition.reduced_composition
        self.bulk_species_symbol = [s.symbol for s in self.bulk_ce.composition.elements]

        # Supplement entries to phase diagram with those from MP database
        if include_mp_entries:
            mpcpa = MPChemPotAnalyzer(
                bulk_ce=self.bulk_ce,
                sub_species=self.sub_species,
                mapi_key=self.mapi_key,
            )
            tempcl = mpcpa.analyze_GGA_chempots(
                full_sub_approach=full_sub_approach
            )  # Use MPentries

            curr_pd = PhaseDiagram(
                list(
                    set().union(
                        mpcpa.entries["bulk_derived"], mpcpa.entries["subs_set"]
                    )
                )
            )
            stable_idlist = {
                i.composition.reduced_composition: [i.energy_per_atom, i.entry_id, i]
                for i in curr_pd.stable_entries
            }
            for mpcomp, mplist in stable_idlist.items():
                matched = False
                for pe in personal_entry_list:
                    if pe.composition.reduced_composition == mpcomp:
                        # #USER: uncomment this if you want additional stable phases of identical
                        # composition included in your phase diagram
                        # if personalentry.energy_per_atom > mplist[0]:
                        #     print('Adding entry from MP-database:',mpcomp,'(entry-id:',mplist[1])
                        #     personal_entry_list.append(mplist[2])
                        matched = True
                if not matched:
                    print(
                        "Adding entry from MP-database:",
                        mpcomp,
                        "(entry-id:",
                        mplist[1],
                    )
                    personal_entry_list.append(mplist[2])
        else:
            # personal_entry_list.append(self.bulk_ce)
            # if you dont have entries for elemental corners of phase diagram then code breaks
            # manually inserting entries with energies of zero for competeness...USER DO NOT USE
            # THIS
            eltcount = {elt: 0 for elt in set(self.bulk_ce.composition.elements)}
            for pentry in personal_entry_list:
                if (
                    pentry.is_element
                    and pentry.composition.elements[0] in eltcount.keys()
                ):
                    eltcount[pentry.composition.elements[0]] += 1
            for elt, eltnum in eltcount.items():
                if not eltnum:
                    s = Structure(
                        [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
                        [elt],
                        [[0, 0, 0]],
                    )
                    eltentry = ComputedStructureEntry(s, 0.0)
                    print(
                        "USER! Note that you have added a fake "
                        + str(elt)
                        + " structure to prevent from breaking the "
                        "Phase Diagram Analyzer.\n As a result DO NOT trust the chemical "
                        "potential results for regions "
                        "of phase diagram that involve the element " + str(elt)
                    )
                    personal_entry_list.append(
                        PDEntry(eltentry.composition, eltentry.energy)
                    )

        personal_entry_list.append(
            PDEntry(
                self.bulk_ce.composition, self.bulk_ce.energy, attribute="Bulk Material"
            )
        )
        unique_entries = list(
            set(personal_entry_list)
        )  # remove duplicate entries if present

        # compute chemical potentials
        if full_sub_approach:
            pd = PhaseDiagram(unique_entries)
            chem_lims = self.get_chempots_from_pd(pd)
        else:
            # first separate out the bulk associated elements from those of substitutional elements
            entry_list = []
            sub_associated_entry_list = []
            for localentry in personal_entry_list:
                bulk_associated = True
                for elt in localentry.composition.elements:
                    if elt not in self.bulk_composition.elements:
                        bulk_associated = False

                if bulk_associated:
                    entry_list.append(localentry)
                else:
                    sub_associated_entry_list.append(localentry)

            # now iterate through and collect chemical potentials
            pd = PhaseDiagram(entry_list)
            chem_lims = self.get_chempots_from_pd(pd)

            finchem_lims = {}  # this will be final chem_lims dictionary
            for key in chem_lims.keys():
                face_list = key.split("-")
                blk, blknom, subnom = self.diff_bulk_sub_phases(face_list)
                finchem_lims[blknom] = {}
                finchem_lims[blknom] = chem_lims[key]

            # Now consider adding single elements to extend the phase diagram,
            # adding new additions to chemical potentials ONLY for the cases
            # where the phases in equilibria are those from the bulk phase
            # diagram. This is essentially the assumption that the majority of
            # the elements in the total composition will be from the native
            # species present rather than the sub species (a good approximation)
            for sub_el in self.sub_species:
                sub_specie_entries = entry_list[:]
                for entry in sub_associated_entry_list:
                    if sub_el in entry.composition.elements:
                        sub_specie_entries.append(entry)

                pd = PhaseDiagram(sub_specie_entries)
                chem_lims = self.get_chempots_from_pd(pd)

                for key in chem_lims.keys():
                    face_list = key.split("-")
                    blk, blknom, subnom = self.diff_bulk_sub_phases(
                        face_list, sub_el=sub_el
                    )
                    # if one less than number of bulk species then can be
                    # grouped with rest of structures
                    if len(blk) == len(self.bulk_species_symbol):
                        if blknom not in finchem_lims.keys():
                            finchem_lims[blknom] = chem_lims[key]
                        else:
                            finchem_lims[blknom][sub_el] = chem_lims[key][sub_el]
                        if "name-append" not in finchem_lims[blknom].keys():
                            finchem_lims[blknom]["name-append"] = subnom
                        else:
                            finchem_lims[blknom]["name-append"] += "-" + subnom
                    else:
                        # if chem pots determined by two (or more) sub-specie
                        # containing phases, skip this facet!
                        continue

            # run a check to make sure all facets dominantly defined by bulk species
            overdependent_chempot = False
            facets_to_delete = []
            for facet_name, cps in finchem_lims.items():
                cp_key_num = (
                    (len(cps.keys()) - 1) if "name-append" in cps else len(cps.keys())
                )
                if cp_key_num != (
                    len(self.bulk_species_symbol) + len(self.sub_species)
                ):
                    facets_to_delete.append(facet_name)
                    print(
                        "Not using facet {} because insufficient number of bulk facets for "
                        "bulk set {} with sub_species set {}. (only dependent on {})."
                        "".format(
                            facet_name,
                            self.bulk_species_symbol,
                            self.sub_species,
                            cps.get("name-append"),
                        )
                    )
            if len(facets_to_delete) == len(finchem_lims):
                overdependent_chempot = True
                print(
                    "Determined chemical potentials to be over dependent"
                    " on a substitutional specie. Needing to revert to full_sub_approach. If "
                    "multiple sub species exist this could take a while/break the code..."
                )
            else:
                finchem_lims = {
                    k: v for k, v in finchem_lims.items() if k not in facets_to_delete
                }

            if not overdependent_chempot:
                chem_lims = {}
                for orig_facet, fc_cp_dict in finchem_lims.items():
                    if "name-append" not in fc_cp_dict:
                        facet_nom = orig_facet
                    else:
                        full_facet_list = orig_facet.split("-")
                        full_facet_list.extend(fc_cp_dict["name-append"].split("-"))
                        full_facet_list.sort()
                        facet_nom = "-".join(full_facet_list)
                    chem_lims[facet_nom] = {
                        k: v for k, v in fc_cp_dict.items() if k != "name-append"
                    }
            else:
                # This is for when overdetermined chempots occur, forcing the full_sub_approach
                # to happen
                for sub, subentries in self.entries["subs_set"].items():
                    for subentry in subentries:
                        entry_list.append(subentry)
                pd = PhaseDiagram(entry_list)
                chem_lims = self.get_chempots_from_pd(pd)

        self.phase_diagram = pd
        chem_lims = {
            "facets": chem_lims,
            "elemental_refs": {
                elt: ent.energy_per_atom
                for elt, ent in self.phase_diagram.el_refs.items()
            },
            "facets_wrt_elt_refs": {},
        }
        for facet, chempot_dict in chem_lims["facets"].items():
            rel_chempot_dict = copy.deepcopy(chempot_dict)
            for elt, chempot_energy in rel_chempot_dict.items():
                rel_chempot_dict[elt] -= chem_lims["elemental_refs"][elt]
            chem_lims["facets_wrt_elt_refs"].update({facet: rel_chempot_dict})
        return chem_lims


class UserChemPotInputGenerator(object):
    """
    For setting up phase diagram for user, based on structures that exist in the MP database
    """

    def __init__(
        self, bulk_composition, sub_species=set(), path_base=".", mapi_key=None
    ):
        """
        Args:
            bulk_composition : Composition of bulk as a pymatgen Composition
                object. This and mapi_key are only actual required input for
                generating set of chemical potentials from Materials Project
                database
            sub_species : set of elemental species that are extrinsic to
                structure defaults to No substitutions needed.
            path_base (str): the base path where the 'PhaseDiagram' folder should be created
                defaults to the local folder
            mapi_key (str): Materials API key to access database
                (if not in ~/.pmgrc.yaml already)
        """
        self.bulk_composition = bulk_composition
        self.bulk_species_symbol = [s.symbol for s in bulk_composition.elements]
        self.redcomp = bulk_composition.reduced_composition
        self.sub_species = sub_species
        self.path_base = path_base
        self.mapi_key = mapi_key
        self.MPC = MPChemPotAnalyzer(sub_species=sub_species)

    def setup_phase_diagram_calculations(
        self,
        full_phase_diagram=False,
        energy_above_hull=0,
        struct_fmt="poscar",
        write_files=False,
        overwrite=False,
        include_elements=True,
        full_sub_approach=True,
    ):
        """
        This method allows for setting up local phase diagram calculations so a user can calculate
        chemical potentials on a level of interest beyond PBE-GGA/GGA+U
        Method is to pull the MP phase diagram and use PBE-GGA level data to decide which phases
        need to be computed

        full_phase_diagram flag has two options:
            False: set up the structures/phases which are stable in GGA phase diagram and are
            relevant for defining
                    the chemical potentials (exist to define the facets adjacent to composition
                    of interest)
            True:  set up the full phase diagram according to all the entries in the MP database
            with elements of interest

        entry_above_hull: allows for a range of energies above hull for each composition being
        set up
                default is 0, meaning just the PBE-GGA ground state phases are set up. If you set
                value to 0.5 then all
                phases within 0.5 eV/atom of PBE-GGA ground state hull will be set up etc.

        struct_fmt: is file format you want structure to be written as. Options are “cif”,
        “poscar”, “cssr”, and “json”
                    Defaults to "poscar"

        write_files: write the calculation folders and structure files to the PhaseDiagram
        folder. Defaults to False.

        overwrite: write files even if PhaseDiagram folder already exists. Defaults to False.

        include_elements: set up all elemental reference phases as well (necessary to properly
            determine absolute chemical potentials, zero-reference elemental energies etc.),
            regardless of whether the elemental phase is a local facet adjacent to the MP
            GGA-calculated stable phase space for the composition of interest. Defaults to True.

        full_sub_approach : Include sub_species in query, sub entries can then be found in
            self.MPC.entries['subs_set']

        """

        # while GGA chem pots won't be used here; use this method for quickly gathering phase
        # diagram object entries AND to find phases of interest if you just want to re-calculate
        # local facets
        if full_sub_approach and self.sub_species:
            MPgga_muvals = self.MPC.get_chempots_from_composition(
                self.bulk_composition, full_sub_approach=True
            )
        else:
            MPgga_muvals = self.MPC.get_chempots_from_composition(self.bulk_composition)

        if full_phase_diagram:
            setupphases = set(
                [localentry.name for localentry in self.MPC.entries["bulk_derived"]]
            )  # all elements in
            # phase diagram
        else:
            if (
                len(self.bulk_composition) == 2
            ):  # neccessary because binary species have chempots
                # written as "A-rich, B-rich"
                setupphases = set(
                    [
                        phase.split("_")[0]
                        for facet in MPgga_muvals.keys()
                        for phase in facet.split("-")
                    ]
                )
            else:
                setupphases = set(
                    [
                        phase
                        for facet in MPgga_muvals.keys()
                        for phase in facet.split("-")
                    ]
                )  # just local facets

            if (
                include_elements
            ):  # add elemental reference phases to structures to setup
                for localentry in self.MPC.entries["bulk_derived"]:
                    if localentry.is_element:
                        setupphases.add(
                            localentry.name
                        )  # set.add() only adds entry if not
                        # already in set

        structures_to_setup = (
            {}
        )  # this will be a list of structure objects which need to be
        # setup locally

        # create phase diagram object for analyzing PBE-GGA energetics of structures computed in
        # MP database
        full_structure_entries = [struct for struct in self.MPC.entries["bulk_derived"]]
        pd = PhaseDiagram(full_structure_entries)

        for entry in full_structure_entries:
            if (entry.name in setupphases) and (
                pd.get_decomp_and_e_above_hull(entry, allow_negative=True)[1]
                <= energy_above_hull
            ):
                with MPRester(api_key=self.mapi_key) as mp:
                    localstruct = mp.get_structure_by_material_id(entry.entry_id)

                # Name to two significant figures
                name = (
                    str(entry.name)
                    + "_EaH="
                    + f"{pd.get_decomp_and_e_above_hull(entry, allow_negative=True)[1]:.2g}"
                )
                if (
                    name in structures_to_setup.keys()
                ):  # Is 2 sig. figures rounding to same
                    # value for two entries?
                    name = (
                        str(entry.name)
                        + "_EaH="
                        + f"{pd.get_decomp_and_e_above_hull(entry, allow_negative=True)[1]:.3g}"
                    )
                try:
                    sg_symbol = localstruct.get_space_group_info()[0]
                except TypeError:
                    try:
                        sg_symbol = localstruct.get_space_group_info(symprec=1e-1)[0]
                    except:
                        try:
                            sg_symbol = localstruct.get_space_group_info(symprec=1e0)[0]
                        except:
                            sg_symbol = None

                structures_to_setup[name] = {
                    "Structure": localstruct,
                    "Energy above Hull": pd.get_decomp_and_e_above_hull(
                        entry, allow_negative=True
                    )[1],
                    "MP Entry ID": entry.entry_id,
                    "Space Group": sg_symbol,
                }

        # Set up structure files locally if desired
        if not write_files:
            print("Returning chempot structures, but not making POSCAR files.")
        else:
            if (
                os.path.exists(os.path.join(self.path_base, "PhaseDiagram"))
                and not overwrite
            ):
                print(
                    "PhaseDiagram folder already exists! Set overwrite = True if you want me to overwrite it"
                )
            else:
                if (
                    os.path.exists(os.path.join(self.path_base, "PhaseDiagram"))
                    and overwrite
                ):
                    print(
                        "PhaseDiagram folder already exists, but f@ck it we're gonna (over)write files"
                    )
                if not os.path.exists(os.path.join(self.path_base, "PhaseDiagram")):
                    os.makedirs(os.path.join(self.path_base, "PhaseDiagram"))
                for localname in structures_to_setup.keys():
                    filename = os.path.join(self.path_base, "PhaseDiagram", localname)
                    if not os.path.isdir(filename):
                        os.makedirs(filename)
                    if struct_fmt == "poscar":
                        outputname = "POSCAR"
                    else:
                        outputname = "structfile"
                    structures_to_setup[localname]["Structure"].to(
                        fmt=struct_fmt, filename=os.path.join(filename, outputname)
                    )
                    # NOTE TO USER. Can use pymatgen here to setup additional calculation
                    # files if interested...

        return structures_to_setup



# coding: utf-8

from __future__ import division

__status__ = "Development"

import os
import tarfile
import unittest
from shutil import copyfile

from monty.json import MontyEncoder
from monty.serialization import dumpfn
from monty.tempfile import ScratchDir
from pymatgen import __file__ as initfilep
from pymatgen.analysis.defects.core import DefectEntry, Substitution, Vacancy
from pymatgen.core import Element, PeriodicSite
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.io.vasp import Locpot, Vasprun
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.core.defects_analyzer import ComputedDefect
from doped.pycdt.utils.parse_calculations import (
    PostProcess,
    SingleDefectParser,
    convert_cd_to_de,
)

pmgtestfiles_loc = os.path.join(
    os.path.split(os.path.split(initfilep)[0])[0], "test_files"
)
file_loc = os.path.abspath(
    os.path.join(__file__, "..", "..", "..", "..", "test_files")
)  # Pycdt Testfiles


class LegacyConversionTest(PymatgenTest):
    def test_convert_cd_to_de(self):
        # create a ComputedDefect object similar to legacy format
        # Vacancy type first
        struc = PymatgenTest.get_structure("VO2")
        struc.make_supercell(3)
        vac = Vacancy(struc, struc.sites[0], charge=-3)
        ids = vac.generate_defect_structure(1)
        defect_data = {"locpot_path": "defect/path/to/files/LOCPOT", "encut": 520}
        bulk_data = {"locpot_path": "bulk/path/to/files/LOCPOT"}

        cse_defect = ComputedStructureEntry(ids, 100.0, data=defect_data)
        cd = ComputedDefect(cse_defect, struc.sites[0], charge=-3, name="Vac_1_O")
        b_cse = ComputedStructureEntry(struc, 10.0, data=bulk_data)

        de = convert_cd_to_de(cd, b_cse)
        self.assertIsInstance(de.defect, Vacancy)
        self.assertIsInstance(de, DefectEntry)
        self.assertEqual(de.parameters["defect_path"], "defect/path/to/files")
        self.assertEqual(de.parameters["bulk_path"], "bulk/path/to/files")
        self.assertEqual(de.parameters["encut"], 520)
        self.assertEqual(de.site.specie.symbol, "O")

        # try again for substitution type
        # (site object had bulk specie for ComputedDefects,
        # while it should have substituional site specie for DefectEntrys...)
        de_site_type = PeriodicSite("Sb", vac.site.frac_coords, struc.lattice)
        sub = Substitution(struc, de_site_type, charge=1)
        ids = sub.generate_defect_structure(1)

        cse_defect = ComputedStructureEntry(ids, 100.0, data=defect_data)
        cd = ComputedDefect(cse_defect, struc.sites[0], charge=1, name="Sub_1_Sb_on_O")

        de = convert_cd_to_de(cd, b_cse)

        self.assertIsInstance(de.defect, Substitution)
        self.assertIsInstance(de, DefectEntry)
        self.assertEqual(de.site.specie.symbol, "Sb")


class SingleDefectParserTest(PymatgenTest):
    def test_all_methods(self):
        # create scratch directory with files....
        # having to do it all at once to minimize amount of time copying over to Scratch Directory
        with ScratchDir("."):
            # setup with fake Locpot object copied over
            copyfile(
                os.path.join(file_loc, "test_path_files.tar.gz"),
                "./test_path_files.tar.gz",
            )
            tar = tarfile.open("test_path_files.tar.gz")
            tar.extractall()
            tar.close()
            blocpot = Locpot.from_file(os.path.join(file_loc, "bLOCPOT.gz"))
            blocpot.write_file("test_path_files/bulk/LOCPOT")
            dlocpot = Locpot.from_file(os.path.join(file_loc, "dLOCPOT.gz"))
            dlocpot.write_file("test_path_files/sub_1_Sb_on_Ga/charge_2/LOCPOT")

            # test_from_path
            sdp = SingleDefectParser.from_paths(
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
                "test_path_files/bulk/",
                18.12,
                2,
            )
            self.assertIsInstance(sdp, SingleDefectParser)
            self.assertIsInstance(sdp.defect_entry.defect, Substitution)
            self.assertEqual(
                sdp.defect_entry.parameters["bulk_path"], "test_path_files/bulk/"
            )
            self.assertEqual(
                sdp.defect_entry.parameters["defect_path"],
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
            )

            # test_freysoldt_loader
            for param_key in [
                "axis_grid",
                "bulk_planar_averages",
                "defect_planar_averages",
                "initial_defect_structure",
                "defect_frac_sc_coords",
            ]:
                self.assertFalse(param_key in sdp.defect_entry.parameters.keys())
            bl = sdp.freysoldt_loader()
            self.assertIsInstance(bl, Locpot)
            for param_key in [
                "axis_grid",
                "bulk_planar_averages",
                "defect_planar_averages",
                "initial_defect_structure",
                "defect_frac_sc_coords",
            ]:
                self.assertTrue(param_key in sdp.defect_entry.parameters.keys())

            # test_kumagai_loader
            for param_key in [
                "bulk_atomic_site_averages",
                "defect_atomic_site_averages",
                "site_matching_indices",
                "sampling_radius",
                "defect_index_sc_coords",
            ]:
                self.assertFalse(param_key in sdp.defect_entry.parameters.keys())
            bo = sdp.kumagai_loader()
            for param_key in [
                "bulk_atomic_site_averages",
                "defect_atomic_site_averages",
                "site_matching_indices",
                "sampling_radius",
                "defect_index_sc_coords",
            ]:
                self.assertTrue(param_key in sdp.defect_entry.parameters.keys())

            # test_get_stdrd_metadata
            sdp.get_stdrd_metadata()
            for param_key in [
                "eigenvalues",
                "kpoint_weights",
                "bulk_energy",
                "final_defect_structure",
                "defect_energy",
                "run_metadata",
            ]:
                self.assertTrue(param_key in sdp.defect_entry.parameters.keys())

            # test_get_bulk_gap_data
            sdp.get_bulk_gap_data()
            self.assertEqual(sdp.defect_entry.parameters["mpid"], "mp-2534")
            self.assertAlmostEqual(sdp.defect_entry.parameters["gap"], 0.1887)

            # test_run_compatibility
            self.assertFalse("is_compatible" in sdp.defect_entry.parameters)
            sdp.run_compatibility()
            self.assertTrue("is_compatible" in sdp.defect_entry.parameters)


class PostProcessTest(PymatgenTest):
    def test_parse_defect_calculations_AND_compile_all(self):
        # testing both parse defect_calculatiosn And the compile all methods because they both require a file structure...
        with ScratchDir("."):
            # make a fake file structure to parse vaspruns and locpot paths
            os.mkdir("bulk")
            copyfile(os.path.join(pmgtestfiles_loc, "vasprun.xml"), "bulk/vasprun.xml")
            os.mkdir(
                "bulk/LOCPOT"
            )  # locpot path just needs to exist..doesnt need to be real locpot file...
            bulktrans = {"supercell": [3, 3, 3], "defect_type": "bulk"}
            dumpfn(bulktrans, "bulk/transformation.json", cls=MontyEncoder)

            os.mkdir("dielectric")
            copyfile(
                os.path.join(pmgtestfiles_loc, "vasprun.xml.dfpt.ionic"),
                "dielectric/vasprun.xml",
            )

            vrobj = Vasprun(os.path.join(pmgtestfiles_loc, "vasprun.xml"))

            os.mkdir("vac_1_As")
            os.mkdir("vac_1_As/charge_0")
            copyfile(
                os.path.join(pmgtestfiles_loc, "vasprun.xml"),
                "vac_1_As/charge_0/vasprun.xml",
            )
            os.mkdir("vac_1_As/charge_0/LOCPOT")  # locpot path just needs to exist
            transchg0 = {
                "charge": 0,
                "supercell": [3, 3, 3],
                "defect_type": "vac_1_As",
                "defect_supercell_site": vrobj.final_structure.sites[0],
            }
            dumpfn(transchg0, "vac_1_As/charge_0/transformation.json", cls=MontyEncoder)

            os.mkdir("vac_1_As/charge_-1")
            copyfile(
                os.path.join(
                    pmgtestfiles_loc, "vasprun.xml.dfpt.unconverged"
                ),  # make this one unconverged...
                "vac_1_As/charge_-1/vasprun.xml",
            )
            os.mkdir("vac_1_As/charge_-1/LOCPOT")  # locpot path just needs to exist
            transchgm1 = {
                "charge": -1,
                "supercell": [3, 3, 3],
                "defect_type": "vac_1_As",
                "defect_supercell_site": vrobj.final_structure.sites[0],
            }
            dumpfn(
                transchgm1, "vac_1_As/charge_-1/transformation.json", cls=MontyEncoder
            )

            os.mkdir("sub_1_Cs_on_As")
            os.mkdir("sub_1_Cs_on_As/charge_2")
            copyfile(
                os.path.join(pmgtestfiles_loc, "vasprun.xml"),
                "sub_1_Cs_on_As/charge_2/vasprun.xml",
            )
            os.mkdir(
                "sub_1_Cs_on_As/charge_2/LOCPOT"
            )  # locpot path just needs to exist
            transchg2 = {
                "charge": 0,
                "supercell": [3, 3, 3],
                "defect_type": "sub_1_Cs_on_As",
                "defect_supercell_site": vrobj.final_structure.sites[1],
                "substitution_specie": "Cs",
            }
            dumpfn(
                transchg2,
                "sub_1_Cs_on_As/charge_2/transformation.json",
                cls=MontyEncoder,
            )

            # now test parse_defect_calculations
            pp = PostProcess(".")
            pdd = pp.parse_defect_calculations()
            self.assertEqual(pdd["bulk_entry"].energy, vrobj.final_energy)
            self.assertEqual(len(pdd["bulk_entry"].structure), 25)
            self.assertEqual(pdd["bulk_entry"].data["bulk_path"], "./bulk")
            self.assertEqual(pdd["bulk_entry"].data["supercell_size"], [3, 3, 3])

            self.assertEqual(len(pdd["defects"]), 2)
            self.assertEqual(pdd["defects"][0].energy, 0.0)
            self.assertEqual(len(pdd["defects"][0].bulk_structure), 25)
            self.assertEqual(
                pdd["defects"][0].parameters["defect_path"], "./vac_1_As/charge_0"
            )
            self.assertEqual(
                pdd["defects"][0].parameters["fldr_name"]
                + "_"
                + str(pdd["defects"][0].charge),
                "vac_1_As_0",
            )
            self.assertEqual(pdd["defects"][0].multiplicity, 1)
            self.assertEqual(
                list(pdd["defects"][0].defect.site.coords),
                list(vrobj.final_structure.sites[0].coords),
            )
            self.assertEqual(pdd["defects"][0].parameters["supercell_size"], [3, 3, 3])

            self.assertEqual(
                list(pdd["defects"][1].defect.site.coords),
                list(vrobj.final_structure.sites[1].coords),
            )
            self.assertEqual(pdd["defects"][1].defect.site.specie.symbol, "Cs")

            # now test compile_all quickly...
            ca = pp.compile_all()
            lk = sorted(list(ca.keys()))
            self.assertEqual(len(lk), 6)
            self.assertEqual(
                lk,
                sorted(["epsilon", "vbm", "gap", "defects", "bulk_entry", "mu_range"]),
            )
            answer = [
                [521.83587174, -0.00263523, 0.0026437],
                [-0.00263523, 24.46276268, 5.381848290000001],
                [0.0026437, 5.381848290000001, 24.42964103],
            ]
            self.assertEqual(ca["epsilon"], answer)
            self.assertEqual(ca["vbm"], 1.5516000000000001)
            self.assertEqual(ca["gap"], 2.5390000000000001)
            self.assertEqual(len(ca["defects"]), 2)
            self.assertEqual(ca["bulk_entry"].energy, vrobj.final_energy)
            # INSERT a simpletest for mu_range...

    def test_get_vbm_bandgap(self):
        with ScratchDir("."):
            os.mkdir("bulk")
            # first check the direct vasprun load method
            copyfile(os.path.join(pmgtestfiles_loc, "vasprun.xml"), "bulk/vasprun.xml")
            pp = PostProcess(".")
            (testvbm, testgap) = pp.get_vbm_bandgap()
            self.assertEqual(testvbm, 1.5516000000000001)
            self.assertEqual(testgap, 2.5390000000000001)
            # secondly check a band gap pull
            pp = PostProcess(".", mpid="mp-2534")  # GaAs mpid
            (testvbm_mpid, testgap_mpid) = pp.get_vbm_bandgap()
            self.assertEqual(testvbm_mpid, 2.6682)
            self.assertEqual(testgap_mpid, 0.18869999999999987)

    def test_get_chempot_limits(self):
        with ScratchDir("."):
            # make a fake file structure to parse vaspruns
            os.mkdir("bulk")
            copyfile(os.path.join(file_loc, "vasprun.xml_GaAs"), "bulk/vasprun.xml")
            pp = PostProcess(".")
            gaas_cp = pp.get_chempot_limits()
            self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(gaas_cp.keys()))
            self.assertEqual(
                [-4.6580705550000001, -4.9884807425],
                [gaas_cp["As-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
            )
            self.assertEqual(
                [-6.609317857500001, -3.03723344],
                [gaas_cp["Ga-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
            )

    def test_dielectric_calculation(self):
        with ScratchDir("."):
            os.mkdir("dielectric")
            copyfile(
                os.path.join(pmgtestfiles_loc, "vasprun.xml.dfpt.ionic"),
                "dielectric/vasprun.xml",
            )
            pp = PostProcess(".")
            eps = pp.parse_dielectric_calculation()
            answer = [
                [521.83587174, -0.00263523, 0.0026437],
                [-0.00263523, 24.46276268, 5.381848290000001],
                [0.0026437, 5.381848290000001, 24.42964103],
            ]
            self.assertEqual(eps, answer)


if __name__ == "__main__":
    unittest.main()
